import os

import numpy as np
import pytest

import gouda
from gouda import GoudaPath

# def test_ensure_dir():
#     test_dir = gouda.ensure_dir("test_dir")
#     test_dir2 = gouda.ensure_dir("test_dir")
#     assert test_dir == test_dir2
#     assert os.path.isdir("test_dir")
#     assert test_dir == "test_dir"
#
#     test_dir_path = gouda.ensure_dir("test_dir", "check1")
#     assert test_dir_path == os.path.join(test_dir, "check1")
#     assert os.path.isdir(test_dir_path)
#
#     pathlib.Path(os.path.join(test_dir_path, 'check2')).touch()
#     with pytest.raises(ValueError):
#         assert gouda.ensure_dir(test_dir_path, "check2")
#
#     # Cleanup
#     os.remove(os.path.join(test_dir_path, 'check2'))
#     os.rmdir(test_dir_path)
#     os.rmdir('test_dir')


def test_init_call():
    test_basic = gouda.GoudaPath(use_absolute=False)
    assert test_basic.path == '.'

    test_abs = GoudaPath('absolute', use_absolute=True)
    test_rel = GoudaPath('relative', use_absolute=False)

    assert os.path.expanduser('~') in test_abs.path
    assert test_abs.path == os.path.abspath('absolute')
    assert test_rel.path == 'relative'
    assert 'relative' in test_rel
    assert 'relative' not in test_abs
    assert test_abs.abspath == os.path.abspath('absolute')
    assert test_rel.abspath == os.path.abspath('relative')
    assert test_abs.basename() == 'absolute'
    assert test_rel.basename() == 'relative'
    assert test_abs.use_absolute
    assert not test_rel.use_absolute

    assert test_abs().path == os.path.abspath('absolute')
    assert test_rel().path == 'relative'
    assert test_abs(use_absolute=False).path == os.path.abspath('absolute')
    assert test_rel(use_absolute=True).path == os.path.abspath('relative')
    assert test_abs(use_absolute=True).path == os.path.abspath('absolute')
    assert test_rel(use_absolute=False).path == test_rel().path

    file_list = ['check1', 'check2', 'check3']
    assert [item.path for item in test_abs(file_list)] == [os.path.join(os.path.abspath('absolute'), item) for item in file_list]
    assert [item.path for item in test_rel(file_list)] == [os.path.join('relative', item) for item in file_list]
    assert [item.path for item in test_abs(*file_list)] == [os.path.join(os.path.abspath('absolute'), item) for item in file_list]
    assert [item.path for item in test_rel(*file_list)] == [os.path.join('relative', item) for item in file_list]

    assert test_abs.abspath == test_abs.realpath
    test_rel.resolve_links()
    assert test_rel.path == test_rel.realpath


def test_representations():
    test_abs = GoudaPath('absolute', use_absolute=True)
    test_rel = GoudaPath('relative', use_absolute=False)

    assert str(test_abs) == os.path.abspath('absolute')
    assert str(test_rel) == 'relative'

    assert repr(test_abs) == "GoudaPath('{}')".format(os.path.abspath('absolute'))
    assert repr(test_rel) == "GoudaPath('{}')".format('relative')

    assert os.fspath(test_abs) == os.path.abspath('absolute')
    assert os.fspath(test_rel) == os.path.abspath('relative')

    assert test_abs.add_basename(test_abs).path == test_abs('absolute').path


def test_navigation():
    test_rel = GoudaPath('relative', use_absolute=False)
    assert test_rel().path == test_rel.basename()
    assert (test_rel / 'check').path == os.path.join('relative', 'check')
    assert (test_rel // 'check').path == os.path.abspath(os.path.join('relative', 'check'))
    assert (test_rel + 'check').path == test_rel.path + 'check'
    assert (test_rel / 'check')[1] == 'check'
    assert (test_rel // 'check')[1] == os.path.abspath(os.path.join('relative', 'check'))[1:].split(os.path.sep)[1]
    assert (test_rel // 'check')[1:2] == os.path.sep.join(os.path.abspath(os.path.join('relative', 'check'))[1:].split(os.path.sep)[1:2])
    assert len(test_rel) == 1
    assert len(test_rel(use_absolute=True)) == len(os.path.abspath('relative')[1:].split(os.path.sep))

    multiples = test_rel(['a', 'b', 'c'])
    assert len(multiples) == 3
    assert multiples[0].path == test_rel('a').path
    assert multiples[2].path == test_rel('c').path
    child_check1 = test_rel('a', use_absolute=False)
    child_check2 = test_rel('a', use_absolute=True)
    assert child_check1.path != child_check2.path
    assert child_check1.abspath == child_check2.path


def test_relation():
    test_dir = GoudaPath('goudapath_test_directory')
    os.mkdir(test_dir)
    os.mkdir(test_dir / 'check_dir1')
    os.mkdir(test_dir / 'check_dir2')
    with open(test_dir / 'check_file.txt', 'w') as _:
        pass

    assert test_dir('check_dir1').exists()
    assert test_dir('check_dir1').is_dir()
    assert GoudaPath('goudapath_test_directory', use_absolute=False).parent_dir().path[:2] == os.pardir
    assert test_dir('check_dir1').parent_dir().path == test_dir.path
    assert not test_dir('doesntexist').exists()
    assert test_dir('check_file.txt').extension() == '.txt'

    children = test_dir.children(dirs_only=True, files_only=False, basenames=False)
    assert len(children) == 2
    assert os.path.join(test_dir.path, 'check_dir1') in [item.path for item in children]
    assert os.path.join(test_dir.path, 'check_dir2') in [item.path for item in children]
    assert os.path.join(test_dir.path, 'check_file.txt') not in [item.path for item in children]

    children = test_dir.children(dirs_only=False, files_only=True, basenames=False)
    assert len(children) == 1
    assert os.path.join(test_dir.path, 'check_dir1') not in [item.path for item in children]
    assert os.path.join(test_dir.path, 'check_file.txt') in [item.path for item in children]

    children = test_dir.children(dirs_only=True, files_only=True, basenames=False)
    assert len(children) == 0

    children = test_dir.children(dirs_only=True, files_only=False, basenames=True, include_hidden=True)
    assert 'check_dir1' in children
    assert 'check_dir2' in children

    assert test_dir.num_children(dirs_only=False, files_only=False, include_hidden=True) == 3
    assert test_dir.num_children(dirs_only=True, files_only=False, include_hidden=False) == 2
    assert test_dir.num_children(dirs_only=False, files_only=True, include_hidden=True) == 1

    with pytest.raises(NotADirectoryError):
        assert test_dir('check_file.txt').children()

    globbed = test_dir.glob('*')
    assert len(globbed) == 3
    assert test_dir('check_dir1').path in globbed

    globbed = test_dir.glob('*.txt', basenames=True)
    assert 'check_file.txt' in globbed

    globbed = test_dir.glob('*', sort=True)
    assert test_dir('check_dir1').path == globbed[0]

    globbed = test_dir.glob('*', sort=True, as_gouda=True)
    assert isinstance(globbed[0], GoudaPath)

    globbed = test_dir.parent_dir().glob('**/*.txt', recursive=True)
    assert test_dir('check_file.txt').path in globbed

    test_img = np.ones([50, 50, 3])
    gouda.image.imwrite(test_dir / 'image1.png', test_img)
    gouda.image.imwrite(test_dir / 'image2.png', test_img)
    image_results = test_dir.get_images()
    assert 'image1.png' in image_results
    assert 'image2.png' in image_results
    assert 'check_file.txt' not in image_results
    assert test_dir('image1.png').is_image()
    with pytest.raises(NotADirectoryError):
        assert test_dir('image1.png').num_children()
    with pytest.raises(NotADirectoryError):
        assert test_dir('image1.png').get_images()
    assert test_dir.is_image() is False
    assert test_dir.num_children(dirs_only=True, files_only=False, include_hidden=True) == 2
    assert test_dir.num_children(dirs_only=False, files_only=True, include_hidden=False) == 3

    os.remove(test_dir / 'check_file.txt')
    os.remove(test_dir / 'image1.png')
    os.remove(test_dir / 'image2.png')
    os.removedirs(test_dir / 'check_dir1')
    os.removedirs(test_dir / 'check_dir2')


def test_encode():
    path = gouda.GoudaPath('.')
    assert os.path.abspath('.').encode() == path.__bytes__()


def test_strings():
    path = gouda.GoudaPath('testerpath', use_absolute=False)
    assert path.startswith('tester')
    assert path.endswith('path')

    path = gouda.GoudaPath('testerpath', use_absolute=True)
    assert path.startswith('tester') is False
    assert path.endswith('path')

    path = gouda.GoudaPath('test  ')
    assert path.rstrip().path == 'test'
    assert path.rstrip('.').path == path.path

    path = gouda.GoudaPath('test.txt')
    assert path[-1] == 'test.txt'
    assert path.basename() == 'test.txt'
    assert path.basicname() == 'test'



"""Binary confusion matrix class"""
import warnings

import colorama
import numpy as np

from .data_methods import num_digits

__author__ = "Sean Mullan"
__copyright__ = "Sean Mullan"
__license__ = "mit"


def underline(string):
    """Shortcut to underline ANSI text"""
    return '\033[4m' + string + '\033[0m'


class BinaryConfusionMatrix(object):
    """
    2D array to represent and evaluate a 2-class confusion matrix.

    Parameters
    ----------
    predictions : type
        Optional initial predictions for the matrix. (the default is None)
    labels : type
        Optional initial lables for the matrix. (the default is None)
    threshold : float
        Threshold to use for binary labels with continuous predictions
    dtype : numpy.dtype
        Numpy variable type to use for the matrix. (the default is np.int)

    Note
    ----
    * Rows represent expected class and columns represent predicted class
    * Matrix is 0-indexed
    * Dtype may be set to change memory usage, but will always be treated as an int. No checking is done to prevent overflow if dtype is manually set.

    """
    def __init__(self, predictions=None, labels=None, threshold=None, dtype=np.int):
        self.threshold = threshold
        self.reset(dtype)
        if predictions is not None:
            self.add(predictions, labels)

    @property
    def dtype(self):
        """The datatype of the values stored in the confusion matrix

        :getter: Return the datatype
        :setter: Re-cast the data in the matrix to a new type
        :type: numpy.dtype
        """
        return self.__matrix.dtype

    @dtype.setter
    def dtype(self, dtype):
        self.__matrix = self.__matrix.astype(dtype)

    @property
    def false_negative(self):
        return self.__matrix[1, 0]

    @property
    def false_positive(self):
        return self.__matrix[0, 1]

    @property
    def matrix(self):
        return self.__matrix

    @property
    def true_negative(self):
        return self.__matrix[0, 0]

    @property
    def true_positive(self):
        return self.__matrix[1, 1]

    def __array__(self, dtype=None):
        if dtype is None:
            return self.__matrix
        else:
            return self.__matrix.astype(dtype)

    def __add__(self, data):
        """Create a new matrix that is the sum of the current one and a new one"""
        new_matrix = self.copy()
        if isinstance(data, BinaryConfusionMatrix):
            new_matrix.add_matrix(data)
        else:
            new_matrix.add(data)
        return new_matrix

    def __getitem__(self, key):
        """Get items from the matrix - Indexing is [expected_class, predicted_class]"""
        return self.__matrix[key]

    def __iadd__(self, data):
        """Add to the current matrix"""
        if isinstance(data, BinaryConfusionMatrix):
            self.add_matrix(data)
        else:
            self.add(data)
        return self

    def __setitem__(self, index, value):
        """Manually set a matrix value"""
        self.__matrix[index] = value

    def __str__(self):
        return str(self.__matrix)

    def __repr__(self):
        digits = max([num_digits(item) for item in np.nditer(self.__matrix)])
        row_str = '{:' + str(digits) + 'd}, {:' + str(digits) + 'd}'
        spacer = " " * 22
        format_string = "BinaryConfusionMatrix([" + row_str + "]\n" + spacer + "[" + row_str + "])"
        return format_string.format(*self.__matrix.flatten())

    def accuracy(self):
        if self.__matrix.sum() == 0:
            return 0
        return (self.__matrix[0, 0] + self.__matrix[1, 1]) / self.__matrix.sum()

    def add(self, predictions, labels=None):
        if labels is None and isinstance(predictions, list):
            # Predictions is actually [predictions, labels]
            if len(predictions) == 2:
                predictions, labels = predictions
            else:
                raise ValueError('If passing a single list, it must be [predictions, labels]')
        elif labels is None:
            predictions = np.array(predictions)
            if 2 not in predictions.shape:
                raise ValueError("At least one dimension must be a stack of prediction/label values".format(predictions.shape))
            axis_to_split = np.where(np.array(predictions.shape) == 2)[0]
            if len(axis_to_split) > 1:
                warnings.warn("Multiple dimensions have size 2. Splitting the first one into prediction/label arrays.", UserWarning)
            predictions, labels = np.split(predictions, 2, axis=axis_to_split[0])
        labels = np.array(labels)
        predictions = np.array(predictions)
        if predictions.max() > 1 or predictions.min() < 0 or labels.max() > 1 or labels.min() < 0:
            raise ValueError("All values must be between 0 and 1")
        labels = np.round(labels).astype(np.int)
        if self.threshold is None:
            predictions = np.round(predictions).astype(np.int)
        else:
            predictions = (predictions > self.threshold).astype(np.int)
        if labels.shape != predictions.shape:
            raise ValueError("Predictions and labels must have the same length/shape")
        labels = np.reshape(labels, [-1])
        predictions = np.reshape(predictions, [-1])
        merged = np.stack([labels, predictions])
        points, counts = np.unique(merged, axis=1, return_counts=True)
        for i in range(points.shape[1]):
            self.__matrix[points[0, i], points[1, i]] += counts[i]

    def add_matrix(self, binary_matrix):
        self.__matrix += binary_matrix.matrix

    def copy(self):
        new_mat = BinaryConfusionMatrix(threshold=self.threshold, dtype=self.dtype)
        new_mat[0, 0] = self.__matrix[0, 0]
        new_mat[0, 1] = self.__matrix[0, 1]
        new_mat[1, 0] = self.__matrix[1, 0]
        new_mat[1, 1] = self.__matrix[1, 1]
        return new_mat

    def count(self):
        """Return the number of items currently in the matrix"""
        return self.__matrix.sum()

    def mcc(self):
        """Return the Matthews correlation coefficient"""
        tp = self.__matrix[1, 1]
        tn = self.__matrix[0, 0]
        fp = self.__matrix[0, 1]
        fn = self.__matrix[1, 0]
        with warnings.catch_warnings(record=False):
            warnings.filterwarnings('error')
            try:
                top = ((tp * tn) - (fp * fn))
                bottom = (tp + fp) * (tp + fn) * (tn + fp) * (tn + fn)
                if bottom == 0:
                    return 0
                result = top / np.sqrt(bottom)
            except RuntimeWarning:
                # Defaults to this in case of overflow issues with large matrices
                n = tn + tp + fn + fp + 0.0
                s = (tp + fn) / n
                p = (tp + fp) / n
                top = (tp / n) - (s * p)
                bottom = p * s * (1 - s) * (1 - p)
                if bottom == 0:
                    return 0
                result = top / np.sqrt(bottom)
        return result

    def precision(self):
        """Return the precision for the true or 1 class [TN / (TP + FP)]"""
        if self.__matrix[:, 1].sum() == 0:
            return 0
        return self.__matrix[1, 1] / self.__matrix[:, 1].sum()

    def reset(self, dtype=None):
        """Reset all matrix entries to 0"""
        if dtype is None:
            dtype = self.__matrix.dtype
        self.__matrix = np.zeros([2, 2], dtype=dtype)

    def sensitivity(self):
        """Return the sensitivity for the true or 1 class [TP / (TP + FN)]"""
        if self.__matrix[1].sum() == 0:
            return 0
        return self.__matrix[1, 1] / self.__matrix[1].sum()

    def specificity(self):
        """Return the specificity for the true or 1 class [TN / (TN + FP)]"""
        if self.__matrix[0].sum() == 0:
            return 0
        return self.__matrix[0, 0] / self.__matrix[0].sum()

    def zero_rule(self):
        """Return the accuracy for guessing the majority class"""
        if self.__matrix.sum() == 0:
            return 0
        return np.max(self.__matrix.sum(axis=1) / self.__matrix.sum())

    def print(self, show_specificity=False, show_sensitivity=False, show_accuracy=False, as_bool=True, return_string=False):
        """Format and print the confusion matrix

        Parameters
        ----------
        show_specificities : bool
            Whether to include specificities at end of columns
        show_sensitivities : bool
            Whether to include sensitivities at end of rows
        as_bool : bool
            Whether to print "True"/"False" instead of 0/1
        return_string : bool
            Whether to return a plain-text version of the matrix (the default is False).

        Returns
        -------
        str
            Confusion matrix formatted for plain-text printing if return_string is True.

        """
        expected_string = u"\u2193" + " Expected"
        predicted_string = u"\u2192" + "  Predicted"
        leading_space = "            "
        confusion_string = "         "
        if self.__matrix.max() == 0:
            item_width = 1
        else:
            item_width = np.ceil(np.log10(self.__matrix.max())).astype(np.int)
        if as_bool:
            item_width = max(item_width, 6)
        item_width = str(item_width)
        if as_bool:
            header_string = "        " + ('| {:^' + item_width + '} ').format('False') + ('| {:^' + item_width + '} ').format("True")
        else:
            header_string = "        " + ('| {:^' + item_width + 'd} ').format(0) + ('| {:^' + item_width + 'd} ').format(1)
        confusion_string += predicted_string + "\n" + expected_string + "  " + underline(header_string) + '\n'
        if as_bool:
            line_string_1 = "  False |"
            line_string_2 = "  True  |"
        else:
            line_string_1 = "    0   |"
            line_string_2 = "    1   |"
        # TN
        line_string_1 += colorama.Fore.GREEN + " {:" + item_width + "d} " + colorama.Style.RESET_ALL + "|"
        # FP
        line_string_1 += " {:" + item_width + "d} |"
        line_string_1 = line_string_1.format(*self.__matrix[0]) + '\n'
        # FN
        line_string_2 += " {:" + item_width + "d} |"
        # TP
        line_string_2 += colorama.Fore.GREEN + " {:" + item_width + "d} " + colorama.Style.RESET_ALL + "|"
        line_string_2 = line_string_2.format(*self.__matrix[1])
        line_string_2 = underline(line_string_2) + '\n'

        confusion_string += leading_space + line_string_1 + leading_space + line_string_2

        if show_accuracy:
            confusion_string += "\nAccuracy:    {:.4f}".format(self.accuracy())
        if show_sensitivity:
            confusion_string += '\nSensitivity: {:.4f}'.format(self.sensitivity())
        if show_specificity:
            confusion_string += '\nSpecificity: {:.4f}'.format(self.specificity())

        print(confusion_string)
        if return_string:
            for item in [colorama.Fore.GREEN, colorama.Style.RESET_ALL, '\033[4m', '\033[0m']:
                confusion_string = confusion_string.replace(item, '')
            return confusion_string



import os

import pytest

import cv2
import numpy as np

from gouda import GoudaPath, image


def test_imwrite_imread():
    image_test = np.ones([100, 100, 3], dtype=np.uint8)
    image_test[:50] *= 0
    image_test[:, :, 2] *= 0
    image_test *= 255

    assert image_test[:, :, 2].sum() == 0
    image.imwrite(GoudaPath('test_RGB.png'), image_test)
    image.imwrite('test_BGR.png', image_test, as_RGB=False)
    image.imwrite('test_singleChannel.png', image_test[:, :, :1])
    image.imwrite('test_2D.png', image_test[:, :, 0])
    image.imwrite('test_uint16.png', image_test.astype(np.uint16))
    with pytest.raises(ValueError):
        assert image.imwrite('failure.png', image_test[:, :, :2])

    assert os.path.isfile('test_RGB.png')
    assert os.path.isfile('test_BGR.png')
    assert os.path.isfile('test_singleChannel.png')
    assert os.path.isfile('test_2D.png')
    assert os.path.isfile('test_uint16.png')
    assert not os.path.isfile('failure.png')

    image_test_in_1 = image.imread(GoudaPath('test_RGB.png'), as_RGB=True, unchanged=False)
    image_test_in_2 = image.imread('test_RGB.png', as_RGB=False, unchanged=False)
    image_test_in_3 = image.imread('test_BGR.png', as_RGB=True, unchanged=False)
    image_test_in_4 = image.imread('test_BGR.png', as_RGB=False, unchanged=False)
    np.testing.assert_array_equal(image_test_in_1, image_test_in_4)
    np.testing.assert_array_equal(image_test_in_2, image_test_in_3)

    image_test_in_5 = image.imread('test_singleChannel.png')
    image_test_in_6 = image.imread('test_singleChannel.png')

    np.testing.assert_array_equal(image_test_in_5, image_test_in_6)

    image_test_in_7 = image.imread('test_uint16.png', unchanged=True)
    assert image_test_in_7.dtype == np.uint16

    image_test_in_8 = image.imread('test_RGB.png', as_greyscale=True, unchanged=False)
    assert image_test_in_8.shape == (100, 100)
    np.testing.assert_allclose(image_test_in_8, cv2.cvtColor(image_test, cv2.COLOR_RGB2GRAY), rtol=0, atol=1)

    os.remove('test_RGB.png')
    os.remove('test_BGR.png')
    os.remove('test_singleChannel.png')
    os.remove('test_2D.png')
    os.remove('test_uint16.png')


def test_rescale():
    image_test = np.ones([10, 10, 3], dtype=np.int)
    image_test[:5] -= 1
    image_test[-2:] += 1
    image_test *= 100
    rescaled_1 = np.unique(image.rescale(image_test))
    rescaled_2 = np.unique(image.rescale(image_test, max_val=2, min_val=0))
    rescaled_3 = np.unique(image.rescale(image_test, max_val=6, min_val=3))
    np.testing.assert_array_equal(rescaled_1, (0, 0.5, 1))
    np.testing.assert_array_equal(rescaled_2, (0, 1, 2))
    np.testing.assert_array_equal(rescaled_3, (3, 4.5, 6))

    image_test_2 = np.ones([10, 10], dtype=np.int)
    rescale_4 = image.rescale(image_test_2, return_type=np.int)
    assert rescale_4.max() == rescale_4.min()
    assert rescale_4[0, 0] == 0
    assert rescale_4.dtype == 'int'
    rescale_5 = image.rescale(image_test_2, return_type=np.float32)
    assert rescale_5.max() == rescale_5.min()
    assert rescale_5[0, 0] == 0
    assert rescale_5.dtype == 'float32'


def test_rescale_columnwise():
    image_test = np.ones([5, 5])
    temp_1 = np.array([0, 0, 2, 4, 4])
    temp_1 = np.stack([temp_1, temp_1], axis=0).transpose([1, 0])
    image_test[:, :2] = image_test[:, :2] * temp_1
    np.testing.assert_array_equal(image_test[:, :2], temp_1)
    temp_2 = np.array([1, 1, 2, 2, 2])
    temp_2 = np.stack([temp_2, temp_2, temp_2], axis=0).transpose([1, 0])
    image_test[:, 2:] *= temp_2
    np.testing.assert_array_equal(image_test[:, 2:], temp_2)
    rescaled_1 = image.rescale(image_test, column_wise=True)
    np.testing.assert_array_equal(image_test, np.array([[0, 0, 1, 1, 1], [0, 0, 1, 1, 1], [2, 2, 2, 2, 2], [4, 4, 2, 2, 2], [4, 4, 2, 2, 2]]))
    np.testing.assert_array_equal(np.unique(rescaled_1[:, :2]), (0, 0.5, 1))
    np.testing.assert_array_equal(np.unique(rescaled_1[:, 2:]), (0, 1))

    image_test_2 = np.ones([5, 5])
    rescaled_2 = image.rescale(image_test_2, column_wise=True)
    assert rescaled_2.max() == rescaled_2.min()
    assert rescaled_2[0, 0] == 0


def test_stack_label():
    test_label = np.ones([10, 10])
    r_label = image.stack_label(test_label, label_channel=0)
    g_label = image.stack_label(test_label, label_channel=1)
    b_label = image.stack_label(test_label, label_channel=2)
    all_label = image.stack_label(test_label, label_channel=-1)
    np.testing.assert_array_equal(r_label + g_label + b_label, all_label)
    np.testing.assert_array_equal(r_label[:, :, 0], g_label[:, :, 1])
    np.testing.assert_array_equal(r_label[:, :, 0], b_label[:, :, 2])
    with pytest.raises(ValueError):
        image.stack_label(test_label, label_channel=4)


def test_laplacian_var():
    image_test = np.zeros([100, 100, 3], dtype=np.uint8)
    assert image.laplacian_var(image_test) == 0

    image_test[:50] += 255
    assert image.laplacian_var(image_test) == 161.3


def test_sobel_var():
    image_test = np.zeros([100, 100, 3], dtype=np.uint8)
    assert image.sobel_var(image_test) == 0

    image_test[:50] += 255
    assert image.sobel_var(image_test) == 629.1456


def test_add_overlay():
    image_1 = np.zeros([100, 100], dtype=np.uint8)
    label_1 = np.zeros([100, 100], dtype=np.uint8)
    label_1[:50] = 255
    overlay1 = image.add_overlay(image_1, label_1)
    np.testing.assert_array_equal(np.unique(overlay1), np.array([0, 128]))

    overlay1b = image.add_overlay(image_1, np.dstack([label_1, np.zeros([100, 100]), np.zeros([100, 100])]))
    np.testing.assert_array_equal(overlay1, overlay1b)

    label_2 = np.copy(label_1).astype(np.float)
    label_2[:50] = 1
    overlay2 = image.add_overlay(image_1[:, :, np.newaxis], label_2)
    np.testing.assert_array_equal(overlay1, overlay2)

    overlay2b = image.add_overlay(image_1.astype(np.float32), label_2.astype(np.float32))
    np.testing.assert_array_equal(np.sign(overlay2) * 0.5, overlay2b)

    image_2 = np.zeros([100, 100, 3], dtype=np.uint8)
    overlay3 = image.add_overlay(image_2, label_2)
    np.testing.assert_array_equal(overlay1, overlay3)

    label_3 = np.copy(label_2)
    label_3[:50] = 255
    overlay4 = image.add_overlay(image_2, label_3)
    np.testing.assert_array_equal(overlay1, overlay4)

    label_4 = np.copy(label_2)
    label_4[50:] = -1
    overlay5 = image.add_overlay(image_2, label_4)
    np.testing.assert_array_equal(overlay5[0, 0], np.array([128, 64, 64]))
    np.testing.assert_array_equal(overlay5[-1, -1], np.array([0, 64, 64]))

    overlay6 = image.add_overlay(image_2, label_4, separate_signs=True)
    np.testing.assert_array_equal(overlay6[0, 0], np.array([0, 128, 0]))
    np.testing.assert_array_equal(overlay6[-1, -1], np.array([128, 0, 0]))




    label_5 = label_3 * 10
    with pytest.warns(UserWarning):
        image.to_uint8(label_5)

    with pytest.raises(ValueError):
        bad_label = np.ones([50, 50])
        image.add_overlay(image_1, bad_label)

    with pytest.raises(ValueError):
        bad_label = np.ones([100, 100, 50])
        image.add_overlay(image_1, bad_label)

    with pytest.raises(ValueError):
        image.split_signs(np.dstack([label_1, label_1, label_1]))




def test_mask_exception():
    image_test = np.ones([100, 100, 3], dtype=np.uint8)
    # image_test[:50] -= 1
    # image_test *= 255
    # label_test = np.ones([100, 100], dtype=np.uint8)
    # label_test[:, :50] -= 1
    # with pytest.raises(ValueError):
    #     assert image.add_overlay(image_test, np.dstack([label_test, label_test]))
    # with pytest.raises(ValueError):
    #     assert image.add_overlay(image_test, label_test * 2)


def test_masked_lineup():
    image_test = np.ones([100, 100, 3], dtype=np.uint8)
    image_test[:50] -= 1
    image_test *= 255
    label_test = np.ones([100, 100], dtype=np.uint8)
    label_test[:, :50] -= 1
    lineup_test = image.masked_lineup(image_test, label_test)
    np.testing.assert_array_equal(lineup_test[0], image_test)
    np.testing.assert_array_equal(lineup_test[1], image.add_overlay(image_test, label_test))
    np.testing.assert_array_equal(lineup_test[2][:, :, 0], label_test * 255)


def test_grabCut():
    image_test = np.ones([100, 100, 3], dtype=np.uint8)
    image_test[:50] -= 1
    image_test *= 255
    label_test = np.ones([100, 100], dtype=np.uint8)
    label_test[:25] = 0
    label_test[25:50] = 2
    label_test[50:75] = 3
    label_test[75:] = 1
    gc_img, gc_mask = image.grabCut(image_test, label_test, iterations=1)
    np.testing.assert_array_equal(gc_img, cv2.bitwise_and(image_test, image_test, mask=gc_mask))

    # assert (cv2.GC_BGD, cv2.GC_PR_BGD, cv2.GC_PR_FGD, cv2.GC_FGD) == 1

    label_test_2 = np.ones([100, 100], dtype=np.float)
    label_test_2[:25] = 0
    label_test_2[25:50] = 0.25
    label_test_2[50:75] = 0.5
    label_test_2[75:] = 0.75
    gc_img_2, gc_mask_2 = image.grabCut(image_test, label_test_2, thresholds=(0.2, 0.4, 0.7), iterations=1)
    np.testing.assert_array_equal(gc_img, gc_img_2)
    np.testing.assert_array_equal(gc_mask, gc_mask_2)
    gc_img_3, gc_mask_3 = image.grabCut(image_test, label_test_2, thresholds=(0.2, 0.4, 0.7), iterations=1, clean=True)
    clean_test = image.clean_grabCut_mask(gc_mask)
    np.testing.assert_array_equal(clean_test, gc_mask_3)
    clean_test_2 = image.clean_grabCut_mask(gc_mask[:, :, np.newaxis])
    np.testing.assert_array_equal(clean_test_2, gc_mask_3[:, :, np.newaxis])


def test_grabCut_exception():
    image_test = np.ones([100, 100, 3], dtype=np.uint8)
    image_test[:50] -= 1
    image_test *= 255
    label_test = np.ones([100, 100], dtype=np.uint8)
    label_test[:50] = 0
    label_test[50:] = 2
    with pytest.raises(ValueError):
        assert image.grabCut(image_test, label_test)
    label_test[:50] = 3
    label_test[50:] = 1
    with pytest.raises(ValueError):
        assert image.grabCut(image_test, label_test)


def test_crop_to_mask():
    image_test = np.ones([100, 100, 3], dtype=np.uint8)
    image_test[:50] = 0
    image_test *= 255
    label_test = np.zeros([100, 100])
    label_test[25:76, 25:76] = 1
    crop_test, mask_test = image.crop_to_mask(image_test, label_test, with_label=True, smoothing=True)
    assert crop_test.shape == (50, 50, 3)
    assert mask_test.shape == (50, 50)
    assert np.all(crop_test[:25] == 0)
    assert np.all(crop_test[25:] == 255)

    crop_test_2 = image.crop_to_mask(image_test, label_test, with_label=False, smoothing=False)
    assert crop_test_2.shape == (50, 50, 3)
    assert crop_test_2.shape == (50, 50, 3)


def test_get_bounds():
    label_test = np.zeros([100, 100])
    label_test[25:76, 25:76] = 1
    (x0, y0), (x1, y1) = image.get_bounds(label_test)
    assert x0 == 25
    assert x1 == 75
    assert y0 == 25
    assert y1 == 75


def test_crop_to_content():
    image_test = np.zeros([100, 100, 3], dtype=np.uint8)
    image_test[25:76, 25:76] = 1
    crop_test = image.crop_to_content(image_test)
    assert crop_test.shape == (50, 50, 3)
    assert crop_test.sum() == crop_test.size


def test_rotate():
    image_test = np.zeros([50, 50, 3], dtype=np.uint8)
    image_test[:25, :25] = 1
    assert image_test.sum() == 25 * 25 * 3
    rotate_test = image.rotate(image_test, degrees=90)
    assert rotate_test.shape == (50, 50, 3)
    assert rotate_test.sum() == image_test.sum()
    assert rotate_test[:25, :25].sum() == 0
    assert rotate_test[:25, 25:].sum() == 25 * 25 * 3

    rotate_test_2 = image.rotate(image_test, degrees=180)
    np.testing.assert_array_equal(rotate_test_2, image_test[::-1, ::-1])

    np.testing.assert_array_equal(image.rotate(image_test, degrees=-90), image.rotate(image_test, degrees=270))
    np.testing.assert_array_equal(image_test, image.rotate(image_test, degrees=360))


def test_rotate_allow_resize():
    image_test = np.zeros([100, 50, 3], dtype=np.uint8)
    rotated_1 = image.rotate(image_test, degrees=90, allow_resize=True)
    assert rotated_1.shape == (50, 100, 3)
    rotated_2 = image.rotate(image_test, degrees=90, allow_resize=False)
    assert rotated_2.shape == (100, 50, 3)


def test_padded_resize():
    image_test = np.ones([50, 50, 3], dtype=np.uint8)
    image.imwrite("test.png", image_test)
    pad_test = image.padded_resize(image_test, size=[50, 50])
    pad_test_file = image.padded_resize("test.png", size=[50, 50])
    assert pad_test.shape == (50, 50, 3)
    np.testing.assert_array_equal(image_test, pad_test)
    np.testing.assert_array_equal(pad_test_file, pad_test)
    os.remove('test.png')

    pad_test_2 = image.padded_resize(image_test, size=[50, 75])
    assert pad_test_2.shape == (50, 75, 3)
    assert pad_test_2[:, :12].sum() == 0
    assert pad_test_2[:, -12:].sum() == 0

    pad_test_3 = image.padded_resize(image_test, size=[75, 75])
    assert pad_test_3.shape == (75, 75, 3)
    assert pad_test_3.sum() == pad_test_3.size

    image_test_2 = np.ones([50, 50, 1], dtype=np.uint8)
    pad_test_4 = image.padded_resize(image_test_2, size=[50, 50])
    np.testing.assert_array_equal(pad_test_4, image_test_2)

    image_test_3 = np.ones([50, 50], dtype=np.uint8)
    pad_test_5 = image.padded_resize(image_test_3, size=[50, 50])
    assert pad_test_5.shape == (50, 50)
    np.testing.assert_array_equal(pad_test_4[:, :, 0], pad_test_5)

    image_test_4 = np.zeros([25, 50, 3], dtype=np.uint8)
    image_test_4[:10] = 1
    pad_test_6 = image.padded_resize(image_test_4, size=(50, 25, 3))
    assert pad_test_6.shape == (50, 25, 3)
    np.testing.assert_array_equal(pad_test_6, image.rotate(image_test_4, degrees=90))

    pad_test_7 = image.padded_resize(image_test, size=[75, 50])
    assert pad_test_7.shape == (75, 50, 3)
    assert pad_test_7[:12].sum() == 0
    assert pad_test_7[-12:].sum() == 0

    pad_test_7 = image.padded_resize(image_test[:, :, 0], size=[75, 50])
    assert pad_test_7.shape == (75, 50)
    assert pad_test_7[:12].sum() == 0
    assert pad_test_7[-12:].sum() == 0


def test_flips():
    image_test = np.ones([50, 50, 3], dtype=np.uint8)
    image_test[:25] = 0
    flip_1 = image.horizontal_flip(image_test)
    np.testing.assert_array_equal(image_test, flip_1)
    flip_2 = image.vertical_flip(image_test)
    assert flip_2[:25].sum() == flip_2[:25].size

    image_test_2 = np.ones([50, 50, 3], dtype=np.uint8)
    image_test_2[:, :25] = 0
    flip_3 = image.horizontal_flip(image_test_2)
    assert flip_3[:, :25].sum() == flip_3[:, :25].size
    flip_4 = image.vertical_flip(image_test_2)
    np.testing.assert_array_equal(flip_4, image_test_2)


def test_adjust_gamma():
    image_test = np.ones([50, 50, 3], dtype=np.uint8) * np.linspace(0, 255, num=50, endpoint=True, dtype=np.uint8).reshape([50, 1, 1])
    assert image_test.dtype == 'uint8'
    gamma_test_1 = image.adjust_gamma(image_test, gamma=2)
    gamma_test_2 = image.adjust_gamma(image_test, gamma=0.5)
    assert gamma_test_1.sum() > image_test.sum()
    assert gamma_test_2.sum() < image_test.sum()


def test_polar_to_cartesian():
    x = np.zeros([200, 200, 3])
    color1 = (255, 0, 0)
    color2 = (255, 255, 0)
    color3 = (255, 0, 255)
    cv2.circle(x, (100, 100), 50, color1, -1)
    cv2.circle(x, (100, 100), 25, color2, -1)
    cv2.circle(x, (100, 100), 10, color3, -1)

    y = image.polar_to_cartesian(x)
    z = image.polar_to_cartesian(x, output_shape=[400, 100, 3])

    assert y.shape == x.shape
    assert z.shape == tuple([400, 100, 3])
    assert (y[0, 0] == color3).all()
    assert (y[50, 0] == color2).all()
    assert (y[100, 0] == color1).all()
    assert (y[0, 0] == z[0, 0]).all()
    assert(y[50, 0] == z[100, 0]).all()
    assert(y[100, 0] == z[200, 0]).all()



"""Classes to handle constant-time mean/stddev updates"""
import numpy as np


class MMean(object):
    """Class to hold a moving mean with constant-time update and memory."""

    def __init__(self):
        self._mean = 0.0
        self._count = 0

    @property
    def count(self):
        """The count of items included in the mean"""
        return self._count

    def __iadd__(self, value):
        """Update the mean, including the given value."""
        self._count += 1
        self._mean += (1.0 / self._count) * (value - self._mean)
        return self

    def __add__(self, value):
        """Add a new value to the mean, does not update class values."""
        return self._mean + value

    def __sub__(self, value):
        """Subtract a new value from the mean, does not update class values."""
        return self._mean - value

    def __str__(self):
        """Return the mean as a string."""
        return str(self._mean)

    def __call__(self):
        """Alternate method for self.mean()"""
        return self._mean

    def mean(self):
        """Return the mean."""
        return self._mean


class MStddev(object):
    """Class to hold a moving standard deviation with constant-time update and memory."""

    def __init__(self):
        self._count = 0.0
        self._mean = 0.0
        self._variance = 0.0

    @property
    def count(self):
        """The count of items included in the standard deviation"""
        return self._count

    def __iadd__(self, value):
        """Update the mean and stddev, including the new value."""
        self._count += 1
        prev_mean = self._mean
        self._mean += (1.0 / self._count) * (value - self._mean)
        self._variance += (value - self._mean) * (value - prev_mean)
        return self

    def __add__(self, value):
        """Add a value to the stddev, does not update class values."""
        return self.stddev() + value

    def __call__(self):
        """Alternate method for self.stddev"""
        return self.stddev()

    def __sub__(self, value):
        """Subtract a value from the stddev, does not update class values."""
        return self.stddev() - value

    def __str__(self):
        """Return the stddev as a string."""
        if self._count == 0:
            return str(0)
        stddev = np.sqrt(self._variance / self._count)
        return str(stddev)

    def mean(self):
        """Return the mean"""
        return self._mean

    def stddev(self):
        """Return the current stddev."""
        if self._count == 0:
            return 0
        return np.sqrt(self._variance / self._count)


class MMeanArray(object):
    """Class to hold an array of element-wise independent means that update in constant-time and memory.

    Note
    ----
        Value shape must be the same or broadcastable to the shape of the
        mean array for all operations.
    """
    def __init__(self, shape, dtype=np.float):
        self._mean = np.zeros(shape, dtype=dtype)
        self._count = 0

    @property
    def shape(self):
        """The shape of the array"""
        return self._mean.shape

    @property
    def dtype(self):
        """The type of data stored in the array"""
        return self._mean.dtype

    @property
    def count(self):
        """The number of examples used for the mean of each item in the array"""
        return self._count

    def __iadd__(self, value):
        """Update the _mean, including the given value."""
        if value.shape != self.shape:
            raise ValueError('Input values must have the same shape as the MMeanArray')
        self._count += 1
        self._mean += (value - self._mean) * (1.0 / self._count)
        return self

    def __add__(self, value):
        """Add a new value to the mean, does not update class values."""
        return self._mean + value

    def __sub__(self, value):
        """Subtract a new value from the mean, does not update class values."""
        return self._mean - value

    def __str__(self):
        """Return the _mean as a string."""
        return str(self._mean)

    def __call__(self):
        """Alternate method for self.mean()"""
        return self._mean

    def mean(self):
        """Return the _mean."""
        return self._mean


class MStddevArray(object):
    """Class to hold an array of element-wise independent standard deviations that update in constant-time and memory.

    Note
    ----
        Value shape must be the same or broadcastable to the shape of the
        mean array for all operations.
    """

    def __init__(self, shape, dtype=np.float):
        self._mean = np.zeros(shape, dtype=dtype)
        self._variance = np.zeros(shape, dtype=dtype)
        self._count = 0

    @property
    def shape(self):
        """The shape of the array"""
        return self._variance.shape

    @property
    def dtype(self):
        """The type of data stored in the array"""
        return self._mean.dtype

    @property
    def count(self):
        """The number of examples used for the standard deviation of each item in the array"""
        return self._count

    def mean(self):
        """Return the mean of the array"""
        return self._mean

    def variance(self):
        """Return the variance of the array"""
        if self._count == 0:
            return np.zeros_like(self._variance)
        return self._variance / self._count

    def __iadd__(self, value):
        """Update the mean and stddev, including the new value."""
        if value.shape != self.shape:
            raise ValueError('Input values must have the same shape as the MStddevArray')
        self._count += 1
        prev_mean = np.copy(self._mean)
        self._mean += (1.0 / self._count) * (value - self._mean)
        self._variance += (value - self._mean) * (value - prev_mean)
        return self

    def __add__(self, value):
        """Add a value to the stddev, does not update class values."""
        return self.stddev() + value

    def __sub__(self, value):
        """Subtract a value from the stddev, does not update class values."""
        return self.stddev() - value

    def __str__(self):
        """Return the stddev as a string."""
        return str(self.stddev())

    def __call__(self):
        """Alternate method for self.stddev()"""
        return self.stddev()

    def stddev(self):
        """Return the current stddev."""
        if self._count == 0:
            return np.zeros_like(self._variance)
        return np.sqrt(self._variance / self._count)



"""Convenience methods to display images using pyplot."""
import matplotlib.pyplot as plt
import numpy as np

__author__ = "Sean Mullan"
__copyright__ = "Sean Mullan"
__license__ = "mit"


def print_grid(*images, figsize=(8, 8), toFile=None, show=True, return_grid_shape=False, **kwargs):
    """Print out images as a grid.

    Parameters
    ----------
    image : list | np.ndarray
        Image(s) to print as a grid
    *images : list | np.ndarray
        More images to print
    figsize : (int, int)
        Figure size to pass to pyplot
    toFile : str
        File to save image to
    show : bool
        Whether to show the grid or not (the default is True)

    Optional Parameters
    -------------------
    Any parameters for matplotlib.pyplot.subplots_adjust can be passed for use in the grid

    Note
    ----
    General accepted formats:
        * List of lists
        * List of dicts with key 'image' with image value
        * List of numpy arrays
        * 2, 3, 4, or 5 dimensional numpy arrays (leading rows will be used as row/column)
    Images can be in shape [x, y] or [x, y, c], but only 1 or 3 channels will work (assumes row/col structure otherwise)
    """
    defaults = ['hspace', 'wspace', 'left', 'bottom', 'right', 'top']
    for item in defaults:
        if item not in kwargs:
            kwargs[item] = None

    if len(images) == 1:
        images = images[0]
    if isinstance(images, (list, tuple)):
        if isinstance(images[0], (list, tuple)):
            # list of lists of images
            rows = len(images)
            cols = max([len(item) for item in images])
            to_show = images
        else:
            # list of images
            rows = 1
            cols = len(images)
            to_show = [images]
    elif isinstance(images, np.ndarray):
        # input as array
        while images.shape[0] == 1:
            images = np.squeeze(images, axis=0)
        if images.shape[-1] == 1:
            images = np.squeeze(images, axis=-1)
        ndim = images.ndim
        if images.shape[-1] == 3:
            ndim -= 1
        if ndim == 2:
            # single image
            rows = 1
            cols = 1
            to_show = [[images]]
        elif ndim == 3:
            rows = 1
            cols = images.shape[0]
            to_show = [[images[i] for i in range(images.shape[0])]]
        elif ndim == 4:
            rows = images.shape[0]
            cols = images.shape[1]
            to_show = [[images[i, j] for j in range(images.shape[1])] for i in range(images.shape[0])]
        else:
            raise ValueError('Invalid array shape: {}'.format(images.shape))
    elif isinstance(images, dict):
        rows = 1
        cols = 1
        to_show = [[images]]
    else:
        raise ValueError("Invalid input type: {}".format(type(images)))

    fig = plt.figure(figsize=figsize)
    gs = fig.add_gridspec(int(rows), int(cols), hspace=kwargs['hspace'], wspace=kwargs['wspace'])

    for row in range(rows):
        image_row = to_show[row]
        for col in range(len(image_row)):
            image = image_row[col]
            if image is None:
                continue
            ax = fig.add_subplot(gs[row, col])

            if isinstance(image, dict):
                image_dict = image
                image = np.squeeze(image_dict['image'])
                if image.ndim == 2:
                    if 'cmap' in image_dict:
                        cmap = image_dict['cmap']
                    else:
                        cmap = 'bone'
                    plt.imshow(image, cmap=cmap)
                else:
                    plt.imshow(image)
                if 'title' in image_dict:
                    ax.set_title(image_dict['title'])
                if 'xlabel' in image_dict:
                    ax.set_xlabel(image_dict['xlabel'])
                if 'ylabel' in image_dict:
                    ax.set_ylabel(image_dict['ylabel'])
            else:
                image = np.squeeze(image)
                plt.imshow(image)
            ax.set_axis_off()
            ax.xaxis.set_major_locator(plt.NullLocator())
            ax.yaxis.set_major_locator(plt.NullLocator())
    plt.subplots_adjust(left=kwargs['left'], bottom=kwargs['bottom'], right=kwargs['right'], top=kwargs['top'])
    if all([kwargs[key] is None for key in defaults]):
        fig.tight_layout()
    if 'suptitle' in kwargs:
        plt.suptitle(kwargs['suptitle'])
    if toFile is not None:
        plt.savefig(toFile, dpi=fig.dpi)
    if show:  # pragma: no cover
        # Check manually
        plt.show()
    if return_grid_shape:
        return rows, cols


def print_image(image, figsize=(8, 6.5), toFile=None, show=True, **kwargs):
    """Similar to pyplot.imshow, but with axes and margins for a single image."""
    defaults = {
        'hspace': 0,
        'wspace': 0,
        'left': 0,
        'bottom': 0,
        'right': 1,
        'top': 1,
        'cmap': 'bone'
    }

    for item in defaults:
        if item not in kwargs:
            kwargs[item] = defaults[item]
    image = np.squeeze(image)
    fig = plt.figure(figsize=figsize)
    if image.ndim == 2:
        cmap = kwargs['cmap']
        plt.imshow(image, cmap=cmap)
    else:
        plt.imshow(image)
    plt.gca().set_axis_off()
    plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
    plt.subplots_adjust(top=kwargs['top'],
                        bottom=kwargs['bottom'],
                        right=kwargs['right'],
                        left=kwargs['left'],
                        hspace=kwargs['hspace'],
                        wspace=kwargs['wspace'])
    plt.margins(0, 0)
    plt.gca().xaxis.set_major_locator(plt.NullLocator())
    plt.gca().yaxis.set_major_locator(plt.NullLocator())
    if toFile is not None:
        plt.savefig(toFile, dpi=fig.dpi)
    if show:  # pragma: no cover
        # Check manually
        plt.show()



import pytest

import numpy as np

import gouda


def test_MMean():
    item = gouda.MMean()
    assert item._mean == 0
    assert item.count == 0
    assert item() == item.mean()

    item += 3
    assert item._mean == 3
    assert item.mean() == item._mean
    assert item.count == 1
    item += 1
    assert item._mean == (3 + 1) / 2
    assert item.mean() == item._mean
    assert item.count == 2
    assert str(item) == str(2.0)
    assert item() == item.mean()

    assert item + 2 == 4
    assert item - 2 == 0


def test_MStddev():
    item = gouda.MStddev()
    assert item.count == 0
    assert item._mean == 0
    assert item.mean() == 0
    assert item.stddev() == 0
    assert str(item) == '0'
    assert item() == item.stddev()

    item += 3
    assert item._mean == 3
    assert item.count == 1
    assert item.stddev() == np.std([3])

    item += 4
    item += 8
    assert item._mean == np.mean([3, 4, 8])
    assert item.count == 3
    assert item.stddev() == np.std([3, 4, 8])
    assert item.mean() == np.mean([3, 4, 8])
    assert item() == item.stddev()

    assert item + 1 == item.stddev() + 1
    assert item - 1 == item.stddev() - 1

    assert str(item) == str(item.stddev())


def test_MMeanArray():
    item = gouda.MMeanArray([3, 3])
    np.testing.assert_array_equal(item._mean, np.zeros([3, 3]))
    assert item.count == 0
    assert item.shape == (3, 3)
    assert item.dtype == 'float'
    np.testing.assert_array_equal(item(), item.mean())

    item += np.ones([3, 3])
    np.testing.assert_array_equal(item._mean, np.ones([3, 3]))
    assert item.count == 1

    with pytest.raises(ValueError):
        assert item.__iadd__(np.ones([4, 3]))

    np.testing.assert_array_equal(item + 1, np.ones([3, 3]) * 2)
    np.testing.assert_array_equal(item - 1, np.zeros([3, 3]))

    assert str(item) == str(np.ones([3, 3]))

    np.testing.assert_array_equal(item.mean(), np.ones([3, 3]))
    item += np.ones([3, 3]) * 3
    item += np.ones([3, 3]) * 16
    np.testing.assert_almost_equal(item.mean()[0, 0], np.mean([1, 3, 16]))
    np.testing.assert_almost_equal(item.mean()[-1, -1], np.mean([1, 3, 16]))
    np.testing.assert_array_equal(item(), item.mean())


def test_MStddevArray():
    item = gouda.MStddevArray([3, 3])
    assert item.count == 0
    assert item.shape == (3, 3)
    assert item.dtype == 'float'
    np.testing.assert_array_equal(item.stddev(), np.zeros([3, 3]))
    np.testing.assert_array_equal(item.mean(), np.zeros([3, 3]))
    np.testing.assert_array_equal(item.variance(), np.zeros([3, 3]))
    np.testing.assert_array_equal(item(), item.stddev())

    item += np.ones([3, 3])
    np.testing.assert_array_equal(item.stddev(), np.zeros([3, 3]))
    assert item.count == 1
    assert str(item) == str(np.zeros([3, 3]))

    with pytest.raises(ValueError):
        assert item.__iadd__(np.ones([4, 3]))

    np.testing.assert_array_equal(item + 1, np.ones([3, 3]))
    np.testing.assert_array_equal(item - 1, np.ones([3, 3]) * -1)

    item += np.ones([3, 3]) * 3
    item += np.ones([3, 3]) * 16
    np.testing.assert_almost_equal(item.stddev()[0, 0], np.std([1, 3, 16]))
    np.testing.assert_almost_equal(item.variance()[0, 0], np.var([1, 3, 16]))
    np.testing.assert_almost_equal(item.mean()[0, 0], np.mean([1, 3, 16]))
    np.testing.assert_array_equal(item(), item.stddev())



import pytest

import numpy as np

from gouda import BinaryConfusionMatrix


def test_init():
    test_mat = BinaryConfusionMatrix()
    np.testing.assert_array_equal(test_mat.matrix, np.array([[0, 0], [0, 0]]))

    test_arr = np.array([[0, 1, 0], [1, 1, 0]])
    test_mat_2 = BinaryConfusionMatrix(test_arr)
    np.testing.assert_array_equal(test_mat_2.matrix, np.array([[1, 0], [1, 1]]))

    assert test_mat_2.dtype == np.int

    test_mat_2.dtype = np.uint8
    assert test_mat_2.dtype == np.uint8

    test_mat_2.reset()
    np.testing.assert_array_equal(test_mat_2.matrix, np.array([[0, 0], [0, 0]]))
    assert test_mat_2.dtype == np.uint8

    test_mat_2.reset(dtype=np.float32)
    assert test_mat_2.dtype == np.float32


def test_add():
    test_mat = BinaryConfusionMatrix()
    np.testing.assert_array_equal(test_mat.matrix, np.array([[0, 0], [0, 0]]))

    test_arr = np.array([[0, 1, 0], [1, 1, 0]])
    test_mat.add(test_arr)
    np.testing.assert_array_equal(test_mat.matrix, np.array([[1, 0], [1, 1]]))

    test_arr2 = np.array([[0, 1], [1, 1], [0, 0]])
    test_mat.add(test_arr2)
    np.testing.assert_array_equal(test_mat.matrix, np.array([[2, 0], [2, 2]]))

    with pytest.raises(ValueError):
        test_mat.add(np.ones([3, 3]))

    test_mat2 = BinaryConfusionMatrix()
    test_mat2.add(1, 1)
    np.testing.assert_array_equal(test_mat2.matrix, np.array([[0, 0], [0, 1]]))

    test_mat3 = test_mat2 + [0, 0]
    np.testing.assert_array_equal(test_mat3.matrix, np.array([[1, 0], [0, 1]]))

    test_mat4 = test_mat3 + test_mat2
    np.testing.assert_array_equal(test_mat4.matrix, np.array([[1, 0], [0, 2]]))
    np.testing.assert_array_equal(test_mat3.matrix, np.array([[1, 0], [0, 1]]))
    np.testing.assert_array_equal(test_mat2.matrix, np.array([[0, 0], [0, 1]]))

    test_mat3 += test_mat2
    np.testing.assert_array_equal(test_mat3.matrix, np.array([[1, 0], [0, 2]]))

    test_mat3 += [0, 0]
    np.testing.assert_array_equal(test_mat3.matrix, np.array([[2, 0], [0, 2]]))

    with pytest.raises(ValueError):
        test_mat.add([1, 2])

    with pytest.raises(ValueError):
        test_mat.add(np.ones(10), np.ones(5))

    with pytest.raises(ValueError):
        test_mat.add(np.ones([3, 3, 3]))

    with pytest.raises(ValueError):
        test_mat.add([1, 2, 3])

    test_mat3.add(predictions=[np.ones(3), [0, 0, 0]], labels=None)
    np.testing.assert_array_equal(test_mat3.matrix, np.array([[2, 3], [0, 2]]))

    test_mat3.threshold = 0.6
    test_mat3.add(predictions=[0.3, 0.5, 0.7], labels=[1, 1, 1])
    np.testing.assert_array_equal(test_mat3.matrix, np.array([[2, 3], [2, 3]]))


def test_parameters():
    test_mat = BinaryConfusionMatrix()
    assert test_mat.print(return_string=True, as_bool=False) == '         →  Predicted\n↓ Expected          | 0 | 1 \n                0   | 0 | 0 |\n                1   | 0 | 0 |\n'

    test_mat.add(np.array([[1, 1, 1, 1], [1, 1, 1, 1]]))
    test_mat.add(np.array([[1, 1, 1], [0, 0, 0]]))
    with pytest.warns(UserWarning):
        test_mat.add(np.array([[0, 0], [0, 0]]))
    test_mat.add(np.array([[0], [1]]))

    assert test_mat.true_positive == 4
    assert test_mat.true_negative == 2
    assert test_mat.false_positive == 3
    assert test_mat.false_negative == 1

    assert test_mat[1, 1] == 4
    assert test_mat[0, 0] == 2

    assert str(test_mat) == '[[2 3]\n [1 4]]'
    assert repr(test_mat) == 'BinaryConfusionMatrix([2, 3]\n                      [1, 4])'
    test_mat.print()
    assert test_mat.print(as_bool=True, return_string=True, show_specificity=True, show_accuracy=True, show_sensitivity=True) == '         →  Predicted\n↓ Expected          | False  |  True  \n              False |      2 |      3 |\n              True  |      1 |      4 |\n\nAccuracy:    0.6000\nSensitivity: 0.8000\nSpecificity: 0.4000'


def test_math():
    test_mat = BinaryConfusionMatrix()
    assert test_mat.mcc() == 0
    assert test_mat.sensitivity() == 0
    assert test_mat.specificity() == 0
    assert test_mat.precision() == 0
    assert test_mat.accuracy() == 0
    assert test_mat.zero_rule() == 0
    test_mat.add(np.array([[1, 1, 1, 1], [1, 1, 1, 1]]))
    test_mat.add(np.array([[1, 1, 1], [0, 0, 0]]))
    with pytest.warns(UserWarning):
        test_mat.add(np.array([[0, 0], [0, 0]]))
    test_mat.add(np.array([[0], [1]]))

    assert test_mat.accuracy() == 6 / 10
    assert test_mat.count() == 10
    assert test_mat.sensitivity() == 4 / 5
    assert test_mat.specificity() == 2 / 5
    assert test_mat.precision() == 4 / 7
    assert test_mat.zero_rule() == 5 / 10

    n = 10
    s = (4 + 1) / n
    p = (4 + 3) / n
    top = (4 / n) - (s * p)
    bottom = p * s * (1 - p) * (1 - s)
    manual_mcc = top / np.sqrt(bottom)

    np.testing.assert_almost_equal(test_mat.mcc(), manual_mcc)

    test_mat2 = BinaryConfusionMatrix()
    test_mat2[1, 1] = 99999999999
    assert test_mat2.mcc() == 0
    test_mat2[0, 0] = 99999999999
    assert test_mat2.mcc() == 1


def test_array():
    test_mat = BinaryConfusionMatrix()
    assert test_mat.__array__().dtype == test_mat.dtype
    assert test_mat.__array__(np.uint8).dtype == np.uint8



"""Confusion matrix class"""
import warnings

import colorama
import numpy as np

__author__ = "Sean Mullan"
__copyright__ = "Sean Mullan"
__license__ = "mit"


def underline(string):
    """Shortcut to underline ANSI text"""
    return '\033[4m' + string + '\033[0m'


class ConfusionMatrix(object):
    """
    2D array to represent and evaluate a confusion matrix.

    Parameters
    ----------
    predictions : type
        Optional initial predictions for the matrix. (the default is None)
    labels : type
        Optional initial lables for the matrix. (the default is None)
    threshold : float
        Threshold to use for binary labels with continuous predictions
    num_classes : int
        Number of classes to use for the matrix
    dtype : numpy.dtype
        Numpy variable type to use for the matrix

    Note
    ----
    * Rows represent expected class and columns represent predicted class
    * Threshold only used for binary class probabilities
    * Matrix is 0-indexed
    * Dtype may be set to change memory usage, but will always be treated as an int. No checking is done to prevent overflow if dtype is manually set.

    """
    def __init__(self, predictions=None, labels=None, threshold=None, num_classes=None, dtype=np.int):

        self.matrix = None
        self._num_classes = 0
        self.threshold = threshold
        self.add_warned = False
        if num_classes is not None:
            self.reset(num_classes, dtype=dtype)
        elif predictions is None and labels is None:
            self.reset(2, dtype=dtype)
        if predictions is not None and labels is not None:
            self.add(predictions, labels, threshold=threshold)
            if self.matrix.dtype != dtype:
                self.matrix = self.matrix.astype(dtype)

    @property
    def shape(self):
        """The shape of the confusion matrix"""
        return self.matrix.shape

    @property
    def size(self):
        """The size of the confusion matrix"""
        return self.matrix.size

    @property
    def dtype(self):
        """The datatype of the values stored in the confusion matrix

        :getter: Return the datatype
        :setter: Re-cast the data in the matrix to a new type
        :type: numpy.dtype
        """
        return self.matrix.dtype

    @dtype.setter
    def dtype(self, dtype):
        self.matrix = self.matrix.astype(dtype)

    @property
    def num_classes(self):
        """Number of classes represented in the confusion matrix"""
        return self._num_classes

    def reset(self, num_classes=None, dtype=None):
        """Reset all matrix entries

        Parameters
        ----------
        num_classes : int
            The number of classes in the new matrix. (defaults to the current matrix size).
        dtype : type
            Numpy type of the new matrix
        """
        if num_classes is None:
            num_classes = self._num_classes
        if num_classes <= 0:
            raise ValueError("Matrix must have at least 1 class")
        if dtype is None:
            dtype = self.matrix.dtype
        self._num_classes = num_classes
        self.matrix = np.zeros((self._num_classes, self._num_classes), dtype=dtype)

    def __iadd__(self, data):
        """Add single datapoint (predicted, expected)"""
        self.add(data[0], data[1])
        return self

    def __add__(self, matrix):
        """Add two matrices together.

        NOTE: Output dtype defaults to first matrix type."""
        incoming_matrix = np.copy(matrix.matrix)
        if self.matrix.dtype != matrix.dtype:
            warnings.warn("Second matrix converted from {} to {} in order to match first matrix.".format(matrix.dtype, self.matrix.dtype), UserWarning)
            incoming_matrix = incoming_matrix.astype(self.matrix.dtype)
        output_size = max(self._num_classes, matrix.num_classes)
        output = np.zeros((output_size, output_size), dtype=self.dtype)
        output[:self._num_classes, :self._num_classes] += self.matrix
        output[:matrix.num_classes, :matrix.num_classes] += incoming_matrix
        output_mat = ConfusionMatrix(num_classes=output_size, dtype=self.matrix.dtype)
        output_mat.matrix = output
        return output_mat

    def __str__(self):
        return str(self.matrix)

    def __getitem__(self, key):
        """Access values of the confusion matrix"""
        return self.matrix[key]

    def __setitem__(self, key, value):
        """Manually set values of the confusion matrix -- NOT RECOMMENDED - USE ADDING/RESET METHODS"""
        self.matrix[key] = value

    def __len__(self):
        """Get length of confusion matrix (number of classes)"""
        return self._num_classes

    def count(self):
        """Count the number of items in the matrix"""
        return self.matrix.sum()

    def accuracy(self):
        """Get the total accuracy in the matrix"""
        return np.sum([self.matrix[i, i] for i in range(self._num_classes)]) / np.sum(self.matrix) if np.sum(self.matrix) > 0 else 0

    def specificity(self, class_index=None):
        """Return the specificity of all classes or a single class.

        NOTE
        ----
        specificity = (true negative) / (true negative + false positive) for each class.
        """
        if class_index is None:
            tn = np.array([sum([self.matrix[j, :i].sum() + self.matrix[j, i + 1:].sum() for j in range(self._num_classes) if j != i]) for i in range(self._num_classes)])
            fp = np.array([self.matrix[i, :].sum() - self.matrix[i, i].sum() for i in range(self._num_classes)])
            return np.divide(tn, tn + fp, where=(tn + fp) > 0)

        else:
            tn = sum([self.matrix[j, :class_index].sum() + self.matrix[j, class_index + 1:].sum() for j in range(self._num_classes) if j != class_index])
            fp = self.matrix[class_index, :].sum() - self.matrix[class_index, class_index].sum()
            return np.divide(tn, tn + fp, where=(tn + fp) > 0)

    def sensitivity(self, class_index=None):
        """Return the sensitivity of all classes or a single class. AKA recall

        NOTE
        ----
        sensitivity = (true positive) / (true positive + false negative) for each class.
        """
        if class_index is None:
            return [self.matrix[i, i] / self.matrix[i, :].sum() if self.matrix[i, :].sum() > 0 else 0 for i in range(self._num_classes)]
        else:
            return self.matrix[class_index, class_index] / self.matrix[class_index, :].sum() if self.matrix[class_index, :].sum() > 0 else 0

    def precision(self, class_index=None):
        """Return the precision of all classes or a single class.

        NOTE
        ----
        precision = (true positive) / (true positive + false positive)
        """
        if class_index is None:
            return [self.matrix[i, i] / self.matrix[:, i].sum() if self.matrix[:, i].sum() > 0 else 0 for i in range(self._num_classes)]
        else:
            return self.matrix[class_index, class_index] / self.matrix[:, class_index].sum() if self.matrix[class_index, :].sum() > 0 else 0

    def mcc(self):
        """Return the Matthews correlation coefficient of a binary confusion matrix.

        NOTE
        ----
        mcc = ((tp * tn) - (fp * fn)) / sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn))
        """
        if self._num_classes != 2:
            raise ValueError("Matthews correlation coefficient only applies to binary classifications")
        tp = self.matrix[1, 1]
        tn = self.matrix[0, 0]
        fp = self.matrix[0, 1]
        fn = self.matrix[1, 0]
        # tn = sum([self.matrix[j, j] for j in range(self._num_classes) if j != 1])
        # fp = sum([self.matrix[j, 1] for j in range(self._num_classes) if j != 1])
        # fn = sum([self.matrix[:, j].sum() - self.matrix[j, j] for j in range(self._num_classes) if j != 1])
        with warnings.catch_warnings(record=False):
            warnings.filterwarnings('error')
            try:
                result = ((tp * tn) - (fp * fn)) / np.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn))
            except RuntimeWarning:
                # Defaults to this in case of overflow issues with large matrices
                n = tn + tp + fn + fp + 0.0
                s = (tp + fn) / n
                p = (tp + fp) / n
                top = (tp / n) - (s * p)
                bottom = p * s * (1 - s) * (1 - p)
                result = top / np.sqrt(bottom)
        return result

    def zero_rule(self):
        """Returns the accuracy as if only the most common class is predicted"""
        return np.max(self.matrix.sum(axis=1) / self.matrix.sum())

    @staticmethod
    def from_array(predicted, expected, threshold=None):
        mat = ConfusionMatrix()
        mat.add_array(predicted, expected, threshold=threshold)
        return mat

    def add_array(self, predicted, expected, threshold=None):
        """Add data to the confusion matrix as numpy arrays
        Parameters
        ----------
        predicted : numpy.ndarray
            Predicted values to add to the matrix either in same shape as expected or with shape [samples, classes] for probabilities
        expected : numpy.ndarray
            Expected values to add to the matrix
        threshold : type
            Threshold to use for predicted probabilities of binary classes. Defaults to self.threshold

        """
        if threshold is None:
            threshold = self.threshold
        if not isinstance(predicted, np.ndarray):
            raise ValueError("predicted and expected must be arrays, not {}".format(type(predicted)))
        if not isinstance(expected, np.ndarray):
            raise ValueError("predicted and expected must be arrays, not {}".format(type(expected)))
        if 'float' in predicted.dtype.name:
            if predicted.ndim == 2 and predicted.shape[1] > 1:
                # Assumes predicted samples as [samples, classes]
                predicted = np.argmax(predicted, axis=1)
            else:
                if threshold is None:
                    predicted = np.round(predicted)
                else:
                    predicted = predicted > threshold
        if 'float' in expected.dtype.name:
            warnings.warn("Float type labels will be automatically rounded to the nearest integer", UserWarning)
            expected = np.round(expected).astype(np.int)
        if not ('int' in expected.dtype.name or 'bool' in expected.dtype.name):
            raise ValueError("Expected must be either an int or a bool, not {}".format(expected.dtype))
        max_in = max(expected.max(), predicted.max()) + 1
        if self.matrix is None:
            self.reset(max_in, dtype=expected.dtype)
        if self._num_classes < max_in:
            new_matrix = np.zeros((max_in, max_in), dtype=self.dtype)
            new_matrix[:self._num_classes, :self._num_classes] += self.matrix
            self.matrix = new_matrix
            self._num_classes = max_in
        expected = expected.astype(self.dtype).flatten()
        predicted = predicted.astype(self.dtype).flatten()
        if expected.shape != predicted.shape:
            raise ValueError("Expected and predicted must have same shape")
        merged = np.stack([expected, predicted])
        points, counts = np.unique(merged, axis=1, return_counts=True)
        for i in range(points.shape[1]):
            self.matrix[points[0, i], points[1, i]] += counts[i]

    def add(self, predicted, expected, threshold=None):
        """Add data to the confusion Matrix

        Parameters
        ----------
        predicted : [bool, float, int, list]
            Predicted value(s) to add to the matrix
        expected : [bool, float, int, list]
            Expected value(s) to add to matrix
        threshold : type
            Threshold used for predicted probabilities of binary classes. Defaults to self.threshold

        NOTE
        ----
        Accepted formats:
            * predicted class     vs expected class
            * probability of true vs expected boolean
            * class probabilities vs expected class
            * list of predictions vs list of class labels (equal length)
        """
        if threshold is None:
            threshold = self.threshold
        if isinstance(predicted, (float, np.float_)):
            # Single value: prediction of True (class 1)
            if self.threshold is not None:
                predicted_class = 1 if predicted > threshold else 0
            else:
                if not self.add_warned:
                    self.add_warned = True
                    print("Warning: Float predicted classes without a threshold are rounded to the nearest integer.")
                predicted_class = np.round(predicted).astype(int)
        elif isinstance(predicted, (bool, np.bool_)):
            predicted_class = 1 if predicted else 0
        elif isinstance(predicted, (int, np.int_)):
            # Single value: class label
            predicted_class = predicted
        elif isinstance(expected, (float, int, bool, np.bool_)) and isinstance(predicted, (list, np.ndarray)):
            # Class probabilities with single expected label
            predicted_class = np.argmax(predicted).astype(int)
        elif isinstance(predicted, (list, np.ndarray)) and isinstance(expected, (list, np.ndarray))and len(predicted) == len(expected):
            # Paired lists
            for x, y in zip(predicted, expected):
                self.add(x, y, threshold=threshold)
            return
        else:
            raise ValueError("Unsupported input format")

        if not isinstance(expected, (float, np.float_, np.int_, int, bool, np.bool_)):
            print(type(expected))
            raise ValueError("Only 1 expected value per prediction is supported")
        if isinstance(expected, (bool, np.bool_)):
            expected_class = 1 if expected else 0
        elif expected % 1 == 0:
            expected_class = np.round(expected).astype(int)
        else:
            raise ValueError("Expected values must be class label integers or boolean")

        max_in = max(predicted_class, expected_class) + 1
        if self.matrix is None:
            self.reset(max_in, dtype=np.array(expected).dtype)
        if self._num_classes < max_in:
            new_matrix = np.zeros((max_in, max_in), dtype=self.dtype)
            new_matrix[:self._num_classes, :self._num_classes] += self.matrix
            self.matrix = new_matrix
            self._num_classes = max_in
        self.matrix[expected_class, predicted_class] += 1

    def print(self, show_specificities=True, show_sensitivities=True, show_accuracy=True, return_string=False):
        """Format and print the confusion matrix

        Parameters
        ----------
        show_specificities : bool
            Whether to include specificities at end of columns
        show_sensitivities : bool
            Whether to include sensitivities at end of rows
        return_string : bool
            Whether to return a plain-text version of the matrix (the default is False).

        Returns
        -------
        str
            Confusion matrix formatted for plain-text printing if return_string is True.

        """
        specificities = self.specificity()
        sensitivities = self.sensitivity()
        expected_string = u"\u2193" + " Expected"
        predicted_string = u"\u2192" + "  Predicted"
        leading_space = "            "
        confusion_string = "         "
        item_width = str(np.ceil(np.log10(self.matrix.max())).astype(np.int))
        header_string = "        " + "".join([('| {:^' + item_width + 'd} ').format(i) for i in range(self._num_classes)])
        if show_sensitivities:
            header_string += "| Sensitivity"
        confusion_string += predicted_string + "\n" + expected_string + "  " + underline(header_string) + '\n'
        for i in range(self._num_classes):
            line_string = "    {:1d}   |"
            for j in range(self._num_classes):
                if i == j:
                    line_string += colorama.Fore.GREEN + " {:" + item_width + "d} " + colorama.Style.RESET_ALL + "|"
                else:
                    line_string += " {:" + item_width + "d} |"
            if show_sensitivities:
                line_string += " {:.4f}"
                line_string = line_string.format(i, *self.matrix[i], sensitivities[i])
            else:
                line_string = line_string.format(i, *self.matrix[i])
            if i == self.num_classes - 1:
                line_string = underline(line_string)
            confusion_string += leading_space + line_string + '\n'

        if show_specificities:
            specificity_string = '        Specificity '
            for _ in range(self._num_classes):
                specificity_string += '| {:>' + item_width + '.4f} '
            confusion_string += (specificity_string + '\n').format(*specificities)

        if show_accuracy:
            confusion_string += "\nAccuracy: {:.4f}".format(self.accuracy())

        print(confusion_string)
        if return_string:
            for item in [colorama.Fore.GREEN, colorama.Style.RESET_ALL, '\033[4m', '\033[0m']:
                confusion_string = confusion_string.replace(item, '')
            return confusion_string



import pytest

import numpy as np

import gouda


def test_num_digits():
    assert gouda.num_digits(5) == 1
    assert gouda.num_digits(9) == 1
    assert gouda.num_digits(10) == 2
    assert gouda.num_digits(99) == 2
    assert gouda.num_digits(100) == 3
    assert gouda.num_digits(0) == 1
    assert gouda.num_digits(-1) == 1
    assert gouda.num_digits(-10) == 2


def test_arr_sample():
    temp_data = np.arange(1, 5)
    resampled = gouda.arr_sample(temp_data, 2)
    np.testing.assert_array_equal(resampled, np.array([1, 3]))
    resampled2 = gouda.arr_sample(temp_data, 0.5)
    np.testing.assert_array_equal(resampled2, np.array([1, 1, 2, 2, 3, 3, 4, 4]))


def test_arr_sample_exception():
    temp_data = np.arange(1, 5).reshape([2, 2])
    with pytest.raises(ValueError):
        assert gouda.arr_sample(temp_data, 1)


def test_sigmoid():
    np.testing.assert_almost_equal(gouda.sigmoid(0), 0.5)
    np.testing.assert_almost_equal(gouda.sigmoid(-100), 0)
    np.testing.assert_almost_equal(gouda.sigmoid(100), 1)


def test_normalize():
    test_data = np.arange(1000).reshape([10, 10, 10])
    normed_1 = gouda.normalize(test_data)
    np.testing.assert_almost_equal(normed_1.std(), 1)
    assert normed_1.mean() == 0

    normed_2 = gouda.normalize(test_data, axis=1)
    np.testing.assert_array_almost_equal(normed_2.mean(axis=1), np.zeros([10, 10]))
    np.testing.assert_equal(normed_2.std(axis=1), np.ones([10, 10]))

    normed_3 = gouda.normalize(test_data, axis=(0, 1))
    np.testing.assert_array_almost_equal(normed_3.mean(axis=(0, 1)), np.zeros([10,]))
    np.testing.assert_equal(normed_3.std(axis=(0, 1)), 1)


def test_rescale():
    test_data = np.arange(100).reshape([10, 10])
    scaled_1 = gouda.rescale(test_data, new_min=0, new_max=1, axis=1)
    manual = (test_data - test_data.min(axis=1, keepdims=True)) / (test_data.max(axis=1, keepdims=True) - test_data.min(axis=1, keepdims=True))
    np.testing.assert_array_equal(scaled_1, manual)

    scaled_2 = gouda.rescale(test_data, new_min=-1, new_max=2)
    assert scaled_2.max() == 2
    assert scaled_2.min() == -1


def test_factors():
    result1 = gouda.factors(100)
    expected1 = set([1, 2, 4, 5, 10, 20, 25, 50, 100])
    assert len(result1.symmetric_difference(expected1)) == 0

    result2 = gouda.factors(6)
    expected2 = set([1, 2, 3, 6])
    assert len(result2.symmetric_difference(expected2)) == 0

    result3 = gouda.factors(7)
    expected3 = set([1, 7])
    assert len(result3.symmetric_difference(expected3)) == 0

    with pytest.raises(ValueError):
        assert gouda.factors(0)

    with pytest.warns(UserWarning):
        result4 = gouda.factors(-1)
        expected4 = set([1])
        assert len(result4.symmetric_difference(expected4)) == 0


def test_prime_factors():
    result1 = gouda.prime_factors(100)
    expected1 = [2, 2, 5, 5]
    assert result1 == expected1

    result2 = gouda.prime_factors(7)
    expected2 = [7]
    assert result2 == expected2

    with pytest.raises(ValueError):
        assert gouda.prime_factors(0)

    with pytest.warns(UserWarning):
        result3 = gouda.prime_factors(-1)
        expected3 = [1]
        assert result3 == expected3


def test_prime_overlap():
    result1 = gouda.prime_overlap(2, 5)
    assert len(result1) == 0

    result2 = gouda.prime_overlap(4, 10)
    assert result2 == [2]

    result3 = gouda.prime_overlap(672, 42)
    assert result3 == [2, 3, 7]

    result4 = gouda.prime_overlap(42, 672)
    assert result4 == [2, 3, 7]

    result5 = gouda.prime_overlap(7, 5)
    assert result5 == []


def test_flip_dict():
    one2one_dict = {'a': 1, 'b': 2, 'c': 3}
    many2one_dict = {'a': 1, 'b': 1, 'c': 2}
    all2one_dict = {'a': 1, 'b': 1, 'c': 1}

    flip1 = gouda.flip_dict(one2one_dict)
    assert flip1 == {1: 'a', 2: 'b', 3: 'c'}

    flip2 = gouda.flip_dict(one2one_dict, unique_items=True)
    assert flip2 == {1: 'a', 2: 'b', 3: 'c'}

    flip3 = gouda.flip_dict(one2one_dict, force_list_values=True)
    assert flip3 == {1: ['a'], 2: ['b'], 3: ['c']}

    flip4 = gouda.flip_dict(many2one_dict)
    assert flip4 == {1: ['a', 'b'], 2: 'c'}

    flip5 = gouda.flip_dict(many2one_dict, unique_items=True)
    assert flip5 == {1: 'b', 2: 'c'}

    flip6 = gouda.flip_dict(many2one_dict, force_list_values=True)
    assert flip6 == {1: ['a', 'b'], 2: ['c']}

    flip7 = gouda.flip_dict(all2one_dict)
    assert flip7 == {1: ['a', 'b', 'c']}


def test_softmax():
    data = np.arange(10).reshape([5, 2])
    assert gouda.softmax(data).sum() == 1
    assert gouda.softmax(data, axis=0).sum() == 2
    assert gouda.softmax(data, axis=1).sum() == 5



"""Methods for working with data and numpy arrays"""
import warnings

import numpy as np


def arr_sample(arr, rate):
    """Return an array linearly sampled from the input array at the given rate.

    Examples
    --------
    * [1, 2, 3, 4] and rate 2   -> [1, 3]
    * [1, 2, 3, 4] and rate 0.5 -> [1, 1, 2, 2, 3, 3, 4, 4]
    """
    if arr.ndim != 1:
        raise ValueError("Only 1d arrays can be sampled from.")
    i = 0
    out = []
    while i < arr.shape[0]:
        out.append(arr[np.floor(i).astype(np.int)])
        i += rate
    return np.array(out)


def factors(x):
    """Return the factors of x.

    Parameters
    ----------
    x : int
        The number to factorize. Must be a non-zero integer

    Returns
    -------
    factors : set
        The set of factors for x
    """
    if x == 0 or x % 1 != 0:
        raise ValueError("Factors can only be found with non-zero integers")
    if x < 0:
        x = np.abs(x)
        warnings.warn("Only positive factors will be returned, but negative numbers have a positive and negative factor for each.", UserWarning)
    factors = set([1, x])
    for i in range(2, int(np.sqrt(x) + 1)):
        if (x / float(i)) == int(x / i):
            factors.add(int(i))
            factors.add(int(x / i))
    return factors


def flip_dict(dict, unique_items=False, force_list_values=False):
    """Swap keys and values in a dictionary

    Parameters
    ----------
    dict: dictionary
        dictionary object to flip
    unique_items: bool
        whether to assume that all items in dict are unique, potential speedup but repeated items will be lost
    force_list_values: bool
        whether to force all items in the result to be lists or to let unique items have unwrapped values. Doesn't apply if unique_items is true.
    """
    if unique_items:
        return {v: k for k, v in dict.items()}
    elif force_list_values:
        new_dict = {}
        for k, v in dict.items():
            if v not in new_dict:
                new_dict[v] = []
            new_dict[v].append(k)
        return new_dict
    else:
        new_dict = {}
        for k, v in dict.items():
            if v in new_dict:
                if isinstance(new_dict[v], list):
                    new_dict[v].append(k)
                else:
                    new_dict[v] = [new_dict[v], k]
            else:
                new_dict[v] = k
        return new_dict


def num_digits(x):
    if x == 0:
        return 1
    return int(np.ceil(np.log10(np.abs(x) + 1)))


def prime_factors(x):
    """Return the prime factorization of x.

    Parameters
    ----------
    x : int
        The number to factorize. Must be a non-zero integer

    Returns
    -------
    prime_factors : list
        The list of prime factors. Repeated factors will occur multiple times in the list.
    """
    if x == 0 or x % 1 != 0:
        raise ValueError("Factors can only be found with non-zero integers")
    if x < 0:
        x = np.abs(x)
        warnings.warn("Only positive factors will be returned, but negative numbers have a positive and negative factor for each.", UserWarning)
    factors = [x]
    prime_factors = []
    while len(factors) > 0:
        check = factors.pop()
        found = False
        for i in range(2, int(np.sqrt(check) + 1)):
            if (check / float(i)) == int(check / i):
                factors.extend([i, int(check / i)])
                found = True
                break
        if not found:
            prime_factors.append(check)
    return sorted(prime_factors)


def prime_overlap(x, y):
    """Return the prime factors x and y have in common.

    Parameters
    ----------
    x : int
        The first number to factorize
    y: int
        The second number to factorize

    Returns
    -------
    overlap : list
        The list of common factors. Repeated factors are included for the number of common repeats.
    """
    fact_x = prime_factors(x)
    fact_y = prime_factors(y)
    overlap = []
    for i in range(len(fact_x)):  # pragma: no branch
        item = fact_x.pop()
        if item in fact_y:
            overlap.append(item)
            fact_y.remove(item)
        if len(fact_x) == 0 or len(fact_y) == 0:
            break
    return sorted(overlap)


def rescale(data, new_min=0, new_max=1, axis=None):
    """Rescales data to have range [new_min, new_max] along axis or axes indicated."""
    data_range = data.max(axis=axis, keepdims=True) - data.min(axis=axis, keepdims=True)
    x = np.divide(data - data.min(axis=axis, keepdims=True), data_range, where=data_range > 0)
    new_range = new_max - new_min
    return (x * new_range) + new_min


def sigmoid(x):
    """Return the sigmoid of the given value/array."""
    return 1.0 / (1.0 + np.exp(-x) + 1e-7)


def softmax(x, axis=None):
    s = np.max(x, axis=axis, keepdims=True)
    e_x = np.exp(x - s)
    div = np.sum(e_x, axis=axis, keepdims=True)
    return np.divide(e_x, div, where=div != 0)


def normalize(data, axis=None):
    """Return data normalized to have zero mean and unit variance along axis or axes indicated."""
    mean = data.mean(axis=axis, keepdims=True)
    stddev = data.std(axis=axis, keepdims=True)
    return np.divide(data - mean, stddev, where=stddev != 0)



"""Path-like class for easier file navigation"""
import glob
import imghdr
import os

__author__ = "Sean Mullan"
__copyright__ = "Sean Mullan"
__license__ = "mit"


class GoudaPath(os.PathLike):
    """Note: requires python 3.6+"""
    def __init__(self, *path, use_absolute=True):
        """PathLike class for easier file traversal

        Parameters
        ----------
        path : str
            One or more strings to join as a filepath
        use_absolute : bool
            Whether to convert the path to the absolute path (the default is True)
        """
        if len(path) == 1 and isinstance(path[0], GoudaPath):
            path = path[0].path
        elif len(path) == 0:
            path = '.'
        else:
            path = os.path.join(*path)

        self.use_absolute = use_absolute
        if use_absolute:
            self.__path = os.path.abspath(path)
        else:
            self.__path = path

    def __call__(self, *path_args, use_absolute=None):
        """Add to the current path.

        Parameters
        ----------
        *path_args : os.PathLike
            One or more strings or PathLike objects to create new GoudaPaths from
        use_absolute : bool
            Whether to convert the path to the absolute path (defaults to the object parameter value)

        Returns
        -------
        path | paths : GoudaPath | list
            Returns one or more GoudaPaths that are children of the current object path
        """
        if use_absolute is None:
            use_absolute = self.use_absolute
        if len(path_args) == 0:
            return GoudaPath(self, use_absolute=use_absolute)

        if isinstance(path_args[0], list) and len(path_args) == 1:
            path_args = path_args[0]

        if len(path_args) == 1:
            return GoudaPath(os.path.join(self.__path, path_args[0]), use_absolute=use_absolute)
        else:
            return [GoudaPath(os.path.join(self.__path, path_args[i]), use_absolute=use_absolute) for i in range(len(path_args))]

    def __str__(self):
        return self.__path

    def __repr__(self):
        return "GoudaPath('{}')".format(self.__path)

    def __bytes__(self):
        return self.abspath.encode()

    def __truediv__(self, path):
        """Shortcut method for __call__ with a single child path and use_absolute set to False"""
        return GoudaPath(self(path), use_absolute=False)

    def __floordiv__(self, path):
        """Shortcut method for __call__ with a single child path and use_absolute set to True"""
        return GoudaPath(self(path), use_absolute=True)

    def __add__(self, path):
        """Appends a path to the end of the current path.

        Note
        ----
        This appends the strings. This does not join the paths, so it is not a child path of the current path
        """
        return GoudaPath(self.__path + path, use_absolute=self.use_absolute)

    def __getitem__(self, index):
        """Get part of the current path hierarchy."""
        split_path = self.__path.split(os.path.sep)
        if len(split_path[0]) == 0:
            split_path = split_path[1:]
        if isinstance(index, slice):
            return os.path.join(*split_path[index])
        return split_path[index]

    def __len__(self):
        """Get the length of the current path hierarchy."""
        split_path = self.__path.split(os.path.sep)
        if len(split_path[0]) == 0:
            return len(split_path) - 1
        else:
            return len(split_path)

    def __contains__(self, value):
        return value in self.__path

    def __fspath__(self):
        """Note: fspath is always absolute path"""
        return os.fspath(os.path.abspath(self.__path))

    def glob(self, pattern, as_gouda=False, basenames=False, recursive=False, sort=False):
        """Make a glob call starting from the current path.

        Parameters
        ----------
        pattern : str
            Pattern to match with the glob
        as_gouda : bool
            Whether to return results as GoudaPath objects (the default is False)
        basenames : bool
            Whether to return only the basenames of results (the default is False)
        recursive : bool
            The setting for the glob recursive argument (the default is False)
        sort : bool
            Whether to sort the results (the default is False)
        """
        results = glob.glob(os.path.join(self.__path, pattern), recursive=recursive)
        if basenames:
            results = [os.path.basename(item) for item in results]
        if sort:
            results = sorted(results)
        if as_gouda:
            results = [GoudaPath(item, use_absolute=self.use_absolute) for item in results]
        return results

    def parent_dir(self):
        """Return the parent directory of the current path."""
        parent_dir = os.path.dirname(self.__path)
        if len(parent_dir) == 0:
            parent_dir = os.path.join(os.pardir, os.path.basename(os.path.abspath(parent_dir)))
        return GoudaPath(parent_dir, use_absolute=self.use_absolute)

    def num_children(self, dirs_only=True, files_only=False, include_hidden=False):
        """If the path is a directory, return a count of the child paths"""
        if not self.is_dir():
            raise NotADirectoryError("Not a directory: {}".format(self.__path))
        children = [self(child) for child in os.listdir(self.__path)]
        if not include_hidden:
            children = [child for child in children if not child.is_hidden()]
        if dirs_only:
            children = [child for child in children if os.path.isdir(child)]
        if files_only:
            children = [child for child in children if os.path.isfile(child)]
        return len(children)

    def children(self, dirs_only=True, files_only=False, basenames=False, include_hidden=False):
        """If the path is a directory, get the child paths contained by it.

        Parameters
        ----------
        dirs_only: bool (default=True)
            Return only child directories
        files_only: bool (default=False)
            Return only child non-directories
        basenames: bool (default=False)
            Return only the basename of the child paths

        Returns
        -------
        list: a list of the children of the current path
        """

        if not self.is_dir():
            raise NotADirectoryError("Not a directory: {}".format(self.__path))
        children = os.listdir(self.__path)
        children = [self(child) for child in children]
        if not include_hidden:
            children = [child for child in children if not child.is_hidden()]
        if dirs_only:
            children = list(filter(lambda x: os.path.isdir(x), children))
        if files_only:
            children = list(filter(lambda x: os.path.isfile(x), children))
        if basenames:
            children = [child.basename() for child in children]
        return children

    def get_images(self):
        """Return all images contained in the directory of the path"""
        if not self.is_dir():
            raise NotADirectoryError("Not a directory: {}".format(self.__path))
        images = []
        for item in os.listdir(self.__path):
            try:
                if imghdr.what(os.path.join(self.__path, item)) is not None:
                    images.append(item)
            except IsADirectoryError:
                continue
        return images

    def resolve_links(self):
        """Resolve any symbolic links in the path using realpath"""
        self.__path = os.path.realpath(self.__path)

    @property
    def path(self):
        return self.__path

    @property
    def abspath(self):
        return os.path.abspath(self.__path)

    @property
    def realpath(self):
        return os.path.realpath(self.__path)

    def basename(self):
        return os.path.basename(self.__path)

    def basicname(self):
        return os.path.splitext(os.path.basename(self.__path))[0]

    def is_dir(self):
        """Check if the path is a directory"""
        return os.path.isdir(self.__path)

    def is_image(self):
        """Check if the path is an image (see imghdr.what for image types)"""
        try:
            return imghdr.what(self.__path) is not None
        except (IsADirectoryError, FileNotFoundError):
            return False

    def is_hidden(self):
        return self.basename().startswith('.')

    def extension(self):
        return '.' + self.__path.rsplit('.', 1)[-1]

    def exists(self):
        return os.path.exists(self)

    def endswith(self, suffix):
        return self.__path.endswith(suffix)

    def rstrip(self, chars=None):
        return GoudaPath(self.__path.rstrip(chars), use_absolute=self.use_absolute)

    def startswith(self, prefix):
        return self.__path.startswith(prefix)

    def add_basename(self, path):
        return self(os.path.basename(path))



import pytest

import numpy as np

import gouda


def test_ConfusionMatrix_init_properties():
    mat = gouda.ConfusionMatrix()
    np.testing.assert_array_equal(mat.matrix, np.zeros([2, 2]))
    assert mat._num_classes == 2

    mat = gouda.ConfusionMatrix(num_classes=2, dtype=np.uint8)
    assert isinstance(mat.matrix, np.ndarray)
    assert mat.shape == mat.matrix.shape
    assert mat.shape == (2, 2)
    assert mat.dtype == mat.matrix.dtype
    assert mat.matrix.dtype == 'uint8'
    assert mat.size == mat.matrix.size
    assert mat.size == 4
    assert mat.num_classes == 2

    mat.dtype = np.float32
    assert mat.matrix.dtype == mat.dtype
    assert mat.dtype == 'float32'


def test_ConfusionMatrix_init_withData():
    predictions = [0, 0, 0, 1, 1, 1, 2, 2, 2]
    labels = [0, 1, 2, 0, 1, 2, 0, 1, 2]
    mat = gouda.ConfusionMatrix(predictions=predictions, labels=labels)
    assert mat.threshold is None
    assert mat.shape == (3, 3)
    assert mat.size == 9
    assert mat.count() == 9
    np.testing.assert_array_equal(mat.matrix, np.ones([3, 3]))


def test_ConfusionMatrix_init_withThreshold():
    predictions = [0.1, 0.2, 0.6, 0.8, 0.9, 1]
    labels = [0, 0, 0, 1, 1, 1]
    mat = gouda.ConfusionMatrix(predictions=predictions, labels=labels, threshold=0.7)
    assert mat.threshold == 0.7
    assert mat.num_classes == 2
    assert mat.shape == (2, 2)
    assert mat.count() == 6
    np.testing.assert_array_equal(mat.matrix, np.array([[3, 0], [0, 3]]))


def test_ConfusionMatrix_reset():
    predictions = [0, 0, 0, 1, 1, 1, 2, 2, 2]
    labels = [0, 1, 2, 0, 1, 2, 0, 1, 2]
    mat = gouda.ConfusionMatrix(predictions=predictions, labels=labels)
    np.testing.assert_array_equal(mat.matrix, np.ones([3, 3]))

    mat.reset()
    assert mat.num_classes == 3
    np.testing.assert_array_equal(mat.matrix, np.zeros([3, 3]))

    mat.reset(num_classes=2)
    assert mat.num_classes == 2
    np.testing.assert_array_equal(mat.matrix, np.zeros([2, 2]))

    with pytest.raises(ValueError):
        assert mat.reset(-1)


def test_ConfusionMatrix_iadd():
    predictions = [0, 0, 0, 1, 1, 1, 2, 2, 2]
    labels = [0, 1, 2, 0, 1, 2, 0, 1, 2]
    mat = gouda.ConfusionMatrix(predictions=predictions, labels=labels)
    np.testing.assert_array_equal(mat.matrix, np.ones([3, 3]))
    mat += [0, 0]
    np.testing.assert_array_equal(mat.matrix, np.array([[2, 1, 1], [1, 1, 1], [1, 1, 1]]))
    mat += [[0.0, 1.0], [0, 0]]
    np.testing.assert_array_equal(mat.matrix, np.array([[3, 2, 1], [1, 1, 1], [1, 1, 1]]))
    assert mat.count() == 12
    mat += [True, False]
    assert mat.count() == 13
    np.testing.assert_array_equal(mat.matrix, np.array([[3, 3, 1], [1, 1, 1], [1, 1, 1]]))

    mat += [[0.1, 0.2, 0.8], 2]
    assert mat.count() == 14
    np.testing.assert_array_equal(mat.matrix, np.array([[3, 3, 1], [1, 1, 1], [1, 1, 2]]))


def test_ConfusionMatrix_iadd_exception():
    predictions = [0, 0, 0, 1, 1, 1, 2, 2, 2]
    labels = [0, 1, 2, 0, 1, 2, 0, 1, 2]
    mat = gouda.ConfusionMatrix(predictions=predictions, labels=labels)
    with pytest.raises(ValueError):
        assert mat.__iadd__(['test', False])

    with pytest.raises(ValueError):
        assert mat.__iadd__([False, 'test'])

    with pytest.raises(ValueError):
        assert mat.__iadd__([False, 1.2])


def test_ConfusionMatrix_add():
    predictions = [0, 0, 0, 1, 1, 1, 2, 2, 2]
    labels = [0, 1, 2, 0, 1, 2, 0, 1, 2]
    mat1 = gouda.ConfusionMatrix(predictions=predictions, labels=labels, dtype=np.int)
    mat2 = gouda.ConfusionMatrix(predictions=predictions, labels=labels, dtype=np.float)
    assert mat2.dtype == 'float'
    np.testing.assert_array_equal(mat1.matrix, np.ones([3, 3]))
    np.testing.assert_array_equal(mat2.matrix, np.ones([3, 3]))

    with pytest.warns(UserWarning):
        mat3 = mat1 + mat2
    assert mat3.dtype == 'int'
    np.testing.assert_array_equal(mat3.matrix, np.ones([3, 3]) * 2)

    assert mat1.dtype == 'int'
    assert mat2.dtype == 'float'
    with pytest.warns(UserWarning):
        mat4 = mat2 + mat1
    assert mat4.dtype == 'float'
    np.testing.assert_array_equal(mat4.matrix, np.ones([3, 3]) * 2)

    mat5 = gouda.ConfusionMatrix(num_classes=4)
    mat6 = mat5 + mat1
    assert mat6.dtype == 'int'
    np.testing.assert_array_equal(mat6.matrix, np.array([[1, 1, 1, 0], [1, 1, 1, 0], [1, 1, 1, 0], [0, 0, 0, 0]]))


def test_ConfusionMatrix_str():
    predictions = [0, 0, 0, 1, 1, 1, 2, 2, 2]
    labels = [0, 1, 2, 0, 1, 2, 0, 1, 2]
    mat1 = gouda.ConfusionMatrix(predictions=predictions, labels=labels, dtype=np.int)
    assert str(mat1) == str(mat1.matrix)


def test_ConfusionMatrix_accuracy():
    predictions = [0, 1, 0, 1, 1, 1, 2, 1, 2]
    labels = [0, 1, 2, 0, 1, 2, 0, 1, 2]
    mat1 = gouda.ConfusionMatrix(predictions=predictions, labels=labels, dtype=np.int)
    assert mat1.accuracy() == 5 / 9


def test_ConfusionMatrix_specificity():
    predictions = [0, 1, 0, 1, 1, 1, 2, 1, 2]
    labels = [0, 1, 2, 0, 1, 2, 0, 1, 2]
    mat1 = gouda.ConfusionMatrix(predictions=predictions, labels=labels, dtype=np.int)
    specs = mat1.specificity()
    for i in range(3):
        assert specs[i] == mat1.specificity(i)
    np.testing.assert_array_equal(specs, np.array([5 / 7, 1, 5 / 7]))


def test_ConfusionMatrix_sensitivity():
    predictions = [0, 1, 0, 1, 1, 1, 2, 1, 2]
    labels = [0, 1, 2, 0, 1, 2, 0, 1, 2]
    mat1 = gouda.ConfusionMatrix(predictions=predictions, labels=labels, dtype=np.int)
    sens = mat1.sensitivity()
    for i in range(3):
        assert sens[i] == mat1.sensitivity(i)
    np.testing.assert_array_equal(sens, np.array([1 / 3, 1, 1 / 3]))


def test_ConfusionMatrix_print():
    # Needs a visual check?
    predictions = [0, 0, 0, 1, 1, 1, 2, 2, 2]
    labels = [0, 1, 2, 0, 1, 2, 0, 1, 2]
    mat1 = gouda.ConfusionMatrix(predictions=predictions, labels=labels, dtype=np.int)
    mat1.print()
    confusion_string = mat1.print(return_string=True)
    assert confusion_string is not None
    assert len(confusion_string) > 10
    assert 'Specificity' in confusion_string
    assert 'Sensitivity' in confusion_string
    assert 'Accuracy' in confusion_string

    confusion_string2 = mat1.print(show_specificities=False, return_string=True)
    assert 'Specificity' not in confusion_string2
    confusion_string3 = mat1.print(show_sensitivities=False, return_string=True)
    assert 'Sensitivity' not in confusion_string3
    confusion_string4 = mat1.print(show_accuracy=False, return_string=True)
    assert 'Accuracy' not in confusion_string4


def test_ConfusionMatrix_from_array_add_array():
    predictions = np.zeros([100])
    labels = np.zeros([100])
    predictions[25:100] = 1
    labels[50:] = 1
    mat = gouda.ConfusionMatrix()
    with pytest.warns(UserWarning):
        mat.add_array(predictions, labels)
    np.testing.assert_array_equal(mat.matrix, np.array([[25, 25], [0, 50]]))

    mat2 = gouda.ConfusionMatrix.from_array(predictions, labels.astype(np.int))
    np.testing.assert_array_equal(mat2.matrix, mat.matrix)

    mat3 = gouda.ConfusionMatrix()
    mat3.matrix = None
    mat3.add_array(predictions, labels.astype(np.int))
    np.testing.assert_array_equal(mat3.matrix, np.array([[25, 25], [0, 50]]))

    mat4 = gouda.ConfusionMatrix(num_classes=1)
    mat4.add_array(predictions, labels.astype(np.int))
    np.testing.assert_array_equal(mat4.matrix, np.array([[25, 25], [0, 50]]))


def test_ConfusionMatrix_add_array_exception():
    predictions = np.zeros([10])
    labels = [0, 0, 0, 0, 0, 1, 1, 1, 1, 1]
    mat = gouda.ConfusionMatrix()
    with pytest.raises(ValueError):
        mat.add_array(predictions, labels)

    predictions = [0, 0, 0, 0, 0, 1, 1, 1, 1, 1]
    labels = np.zeros([10])
    mat = gouda.ConfusionMatrix()
    with pytest.raises(ValueError):
        mat.add_array(predictions, labels)

    predictions = np.zeros([10])
    labels = np.zeros([10]).astype(np.object)
    mat = gouda.ConfusionMatrix()
    with pytest.raises(ValueError):
        mat.add_array(predictions, labels)

    predictions = np.zeros([10], dtype=np.int)
    labels = np.zeros([8], dtype=np.int)
    with pytest.raises(ValueError):
        mat = gouda.ConfusionMatrix.from_array(predictions, labels)


def test_ConfusionMatrix_add_array_argmax():
    predictions = np.array([[0.1, 0.9], [0.3, 0.7], [0.6, 0.4], [1.0, 0.0]])
    labels = np.array([1, 1, 0, 1])
    mat = gouda.ConfusionMatrix.from_array(predictions, labels)
    np.testing.assert_array_equal(mat.matrix, np.array([[1, 0], [1, 2]]))


def test_ConfusionMatrix_add_array_threshold():
    predictions = np.array([0.1, 0.2, 0.6, 0.8])
    labels = np.array([0, 1, 0, 1])
    mat = gouda.ConfusionMatrix.from_array(predictions, labels)
    np.testing.assert_array_equal(mat.matrix, np.ones([2, 2]))

    predictions = np.array([0.1, 0.2, 0.6, 0.8])
    labels = np.array([0, 1, 0, 1])
    mat = gouda.ConfusionMatrix.from_array(predictions, labels, threshold=0.7)
    np.testing.assert_array_equal(mat.matrix, np.array([[2, 0], [1, 1]]))


def test_ConfusionMatrix_precision_mcc():
    predictions = np.array([0, 0, 1, 1])
    labels = np.array([0, 1, 0, 1])
    mat = gouda.ConfusionMatrix.from_array(predictions, labels)
    np.testing.assert_array_equal(mat.precision(), np.array([0.5, 0.5]))
    np.testing.assert_array_equal(np.array([mat.precision(0), mat.precision(1)]), mat.precision())
    assert mat.mcc() == 0.0

    predictions = np.array([0, 0, 1, 1, 1])
    labels = np.array([0, 0, 0, 1, 1])
    mat2 = gouda.ConfusionMatrix.from_array(predictions, labels)
    np.testing.assert_array_equal(mat2.precision(), np.array([2.0 / 2.0, 2.0 / 3.0]))
    assert mat2.mcc() == 2.0 / 3.0

    mat3 = gouda.ConfusionMatrix.from_array(np.array([1, 2, 3]), np.array([1, 2, 3]))
    with pytest.raises(ValueError):
        mat3.mcc()



"""General file method and JSON handling"""
import copy
import json
import imghdr
import os
import warnings

import numpy as np

from .goudapath import GoudaPath

__author__ = "Sean Mullan"
__copyright__ = "Sean Mullan"
__license__ = "mit"


def ensure_dir(*paths):
    """Check if a given directory exists, and create it if it doesn't. Multiple directories can be passed as a top-to-bottom path structure.

    Parameters
    ----------
    *paths : iterable
        One or more nested directories to ensure

    Returns
    -------
    str
        Joined filepath of all ensured directories

    """
    full_path = ''
    for path in paths:
        if isinstance(path, GoudaPath):
            full_path = os.path.join(full_path, path.path)
        else:
            full_path = os.path.join(full_path, path)
        if os.path.exists(full_path) and os.path.isdir(full_path):
            continue
        elif os.path.exists(full_path):
            raise ValueError("A file without an extension is blocking directory creation at {}".format(full_path))
        else:
            os.makedirs(full_path, exist_ok=True)
    if isinstance(paths[0], GoudaPath):
        return GoudaPath(full_path, use_absolute=paths[0].use_absolute)
    return full_path


def next_filename(filename):
    """Check if a given file exists, and return a new filename for a numbered copy if it does."""
    if os.path.isfile(filename):
        base, extension = filename.rsplit('.', 1)
        i = 2
        while True:
            next_check = '{}_{}.{}'.format(base, i, extension)
            if os.path.isfile(next_check):
                i += 1
            else:
                return next_check
    else:
        return filename


# JSON methods
def load_json(filename):
    """Load a JSON file, and re-form any numpy arrays if :func:`~gouda.save_json` was used to write them."""
    with open(filename, 'r') as f:
        data = json.load(f)
    if isinstance(data, dict) and 'slice_start' in data:
        data = slice(data['slice_start'], data['slice_stop'], data['slice_step'])
    elif isinstance(data, list):
        if data[-1] == 'numpy':
            np_filename = filename.rsplit('.', 1)[0] + '_array.npz'
            arrays = np.load(np_filename)
            data = data[0]
        elif data[-1] == 'numpy_zip':
            np_filename = filename.rsplit('.', 1)[0] + '_arrayzip.npz'
            arrays = np.load(np_filename)
            data = data[0]
        elif data[-1] == 'numpy_embed':
            data = data[0]
        # else:
        #     return data

        def renumpy(_data):
            if isinstance(_data, list):
                for i in range(len(_data)):
                    print('here')
                    _data[i] = renumpy(_data[i])
            elif isinstance(_data, dict):
                if 'numpy_array' in _data:
                    if isinstance(_data['numpy_array'], list):
                        new_data = np.array(_data['numpy_array'], dtype=_data['dtype']).reshape(_data['shape'])
                    else:
                        new_data = arrays[_data['numpy_array']]
                        if new_data.dtype != _data['dtype'] or list(new_data.shape) != _data['shape']:
                            raise ValueError("Numpy array file doesn't match expected stored numpy array data")
                    return new_data
                elif 'slice_start' in _data:
                    _data = slice(_data['slice_start'], _data['slice_stop'], _data['slice_step'])
                else:
                    for key in _data.keys():
                        _data[key] = renumpy(_data[key])
            return _data

        # if len(data) == 1:
        #     data = data[0]
        data = renumpy(data)
    return data


def save_json(data, filename, embed_arrays=True, compressed=False):
    """Save a list/dict/numpy.ndarray as a JSON file.

    Parameters
    ----------
    data : [list, dict, numpy.ndarray]
        Data to save as a JSON file
    filename : string
        Path to write the JSON to
    embed_arrays : bool [defaults to True]
        Whether to embed any numpy arrays into the JSON as lists with metadata. If false saves them to a separate file with placeholders in the JSON.
    compressed : bool [defaults to False]
        If saving numpy arrays in a separate file, this determines if they are zipped or not.

    NOTE
    ----
    JSON files saved this way can be read with any JSON reader, but will have an extra numpy tag at the end that is used to tell :func:`~gouda.load_json` how to read the arrays back in.
    """
    out_arrays = {}
    used_numpy = [False]
    if embed_arrays and compressed:
        warnings.warn('Cannot compress an array that is embedded in a JSON', UserWarning)
        compressed = False

    def unnumpy(_data):
        if isinstance(_data, list):
            new_data = []
            for i in range(len(_data)):
                new_data.append(unnumpy(_data[i]))
        elif isinstance(_data, dict):
            new_data = {}
            for key in _data.keys():
                new_data[key] = unnumpy(_data[key])
        elif isinstance(_data, slice):
            new_data = {'slice_start': _data.start, 'slice_stop': _data.stop, 'slice_step': _data.step}
        elif isinstance(_data, np.ndarray):
            used_numpy[0] = True
            if embed_arrays:
                new_data = {"numpy_array": _data.tolist(), "dtype": str(_data.dtype), "shape": _data.shape}
            else:
                new_data = {"numpy_array": 'array_{}'.format(len(out_arrays)), 'dtype': str(_data.dtype), 'shape': _data.shape}
                out_arrays['array_{}'.format(len(out_arrays))] = _data
        else:
            new_data = copy.copy(_data)
        return new_data

    data = unnumpy(data)
    if used_numpy[0]:
        data = [data]
        if compressed:
            data.append('numpy_zip')
        elif embed_arrays:
            data.append('numpy_embed')
        else:
            data.append('numpy')
    with open(filename, 'w') as f:
        json.dump(data, f, indent=2, sort_keys=True)
    if len(out_arrays) != 0:
        np_filename = filename.rsplit('.', 1)[0]
        if compressed:
            np.savez_compressed(np_filename + '_arrayzip.npz', **out_arrays)
        else:
            np.savez(np_filename + '_array.npz', **out_arrays)


def is_image(path):
    if isinstance(path, GoudaPath):
        return path.is_image()
    else:
        try:
            return imghdr.what(path) is not None
        except IsADirectoryError:
            return False



"""Methods/Shortcuts for modifying and handling image data."""
import warnings

import cv2
import numpy as np

from .goudapath import GoudaPath

__author__ = "Sean Mullan"
__copyright__ = "Sean Mullan"
__license__ = "mit"


def imread(path, as_RGB=False, as_greyscale=False, unchanged=True):
    """SHORTCUT: Load an image from a path using OpenCV modified for RGB.

    Parameters
    ----------
    path: string
        Path to image file
    as_RGB: bool
        If true, assumes an RGB image and flips the channels. (OpenCV uses BGR by default) (the default is True)
    as_greyscale: bool
        If true, reads the image as greyscale (the default is False)
    unchanged: bool
        If true, reads in the image without constraining to 3-channel uint8 format. (the default is True)

    Note
    ----
    * Priority for options is: as_greyscale > as_RGB > unchanged > default
    * Greyscale transforms input image based on perceived color. See [https://docs.opencv.org/2.4/modules/imgproc/doc/miscellaneous_transformations.html#cvtcolor]
    """
    if isinstance(path, GoudaPath):
        path = path.path
    if as_greyscale:
        return cv2.imread(path, 0)
    elif as_RGB:
        return cv2.imread(path)[:, :, ::-1]
    elif unchanged:
        return cv2.imread(path, -1)
    else:
        return cv2.imread(path)


def imwrite(path, image, as_RGB=True):
    """SHORTCUT: Write an image to a path using OpenCV modified for RGB.

    Parameters
    ----------
    path: string
        Path to save image file to
    image: image data
        image data to save - must be uint8/uint16 and with shape [x, y], [x, y, 1], or [x, y, 3]
    as_RGB: bool
        If true, flips the channels before saving (OpenCV assumes BGR image by default)
        """
    if isinstance(path, GoudaPath):
        path = path.path
    if image.ndim == 2:
        image = image[:, :, np.newaxis]
        as_RGB = False
    elif image.ndim == 3 and image.shape[2] == 1:
        as_RGB = False
    elif image.ndim == 3 and image.shape[2] == 3:
        pass
    else:
        raise ValueError("Image must be of shape [x, y, 1], [x, y, 3], or [x, y], not {}".format(image.shape))
    if as_RGB:
        cv2.imwrite(path, image[:, :, ::-1])
    else:
        cv2.imwrite(path, image)


def to_uint8(x):
    """Convert an image to a uint8 type with range [0, 255] based on inferred normalization type."""
    if x.dtype == np.uint8:
        return x
    if x.max() > 1 and x.max() <= 255 and x.min() >= 0:  # input range [0, 255]
        pass
    elif x.min() < 0 and x.min() >= -1 and x.max() <= 1:  # input range [-1, 1]
        x = ((x * 127.5) + 127.5)
    elif x.min() >= 0 and x.max() <= 1:  # input range [0, 1]
        x = (x * 255.0)
    else:
        warnings.warn("Cannot determine input range. Rescaling to [0, 1]")
        x = rescale(x, min_val=0, max_val=255)
    return x.astype(np.uint8)


def rescale(data, column_wise=False, max_val=1, min_val=0, return_type=np.float):
    """Scale either an image or rows of data to [0, 1].

        NOTE
        ----
        If the range of values in a column/image is 0, data will scale to 0
    """
    data = data.astype(np.float)
    if column_wise:
        range = data.max(axis=0) - data.min(axis=0)
        range[range == 0] = 1
        rescaled = (data - data.min(axis=0)) / range
        rescaled[range == 0] = 0
    else:
        range = data.max() - data.min() + 0.0
        if range == 0:
            rescaled = np.zeros_like(data)
        else:
            rescaled = (data - data.min()) / range
    return ((rescaled * (max_val - min_val)) + min_val).astype(return_type)


def stack_label(label, label_channel=0, as_uint8=True):
    """Convert 2d label to 3d.

    Parameters
    ----------
    label : np.ndarray
        2D label to stack
    label_channel : int
        The color channel for the label, or -1 to set all channels to the label (the default is 0)
    as_uint8 : bool
        Whether to convert the label to a uint8
    """
    label = np.squeeze(label)
    if as_uint8:
        label = to_uint8(label)
    if label_channel < 0:
        return np.dstack([label, label, label])
    elif label_channel < 3:
        to_stack = [np.zeros_like(label), np.zeros_like(label), np.zeros_like(label)]
        to_stack[label_channel] = label
        return np.dstack(to_stack)
    else:
        raise ValueError("Not a valid color channel index: {}".format(label_channel))


def laplacian_var(image):
    """Return the laplacian variance of an image"""
    # Laplacian is the rate of change of pixel intensity (2nd order derivative)
    blur = cv2.GaussianBlur(image, (3, 3), 0, 0)
    grey = cv2.cvtColor(blur, cv2.COLOR_BGR2GRAY)
    return cv2.Laplacian(grey, cv2.CV_16S).var()


def sobel_var(image):
    """Return the sobal variance of an image"""
    # Sobel is the gradient of pixel intensity (1st order derivative)
    blur = cv2.GaussianBlur(image, (3, 3), 0, 0)
    grey = cv2.cvtColor(blur, cv2.COLOR_BGR2GRAY)
    grad_x = cv2.Sobel(grey, cv2.CV_16S, 1, 0, 3)
    grad_x = cv2.convertScaleAbs(grad_x)
    grad_y = cv2.Sobel(grey, cv2.CV_16S, 0, 1, 3)
    grad_y = cv2.convertScaleAbs(grad_y)
    return cv2.addWeighted(grad_x, 0.5, grad_y, 0.5, 0).var()


def split_signs(mask):
    """Split a single channel image mask with +/- values into green/red color channels"""
    mask = np.squeeze(mask)
    if mask.ndim != 2:
        raise ValueError("Single channel mask required for split_signs")
    negative = np.clip(mask, -np.inf, 0) * -1
    positive = np.clip(mask, 0, np.inf)
    new_mask = np.dstack([negative, positive, np.zeros_like(positive)])
    return new_mask


def add_overlay(image, mask, label_channel=0, separate_signs=False):
    """Return image with a mask overlay.

    Parameters
    ----------
        image: numpy.ndarray
            Image array with shape (x, y, channels)
        mask: numpy.ndarray
            np array with shape (x, y) and range [0, 1]
        label_channel : int
            The color channel to use for the overlay if separate_signs is False (the default is 0)
        separate_signs : bool
            Whether to separate +/- values of the mask into green/red channels
    """
    output = np.squeeze(image)
    mask = np.squeeze(mask)

    if mask.shape[:2] != image.shape[:2]:
        raise ValueError('Mask width/height does not match image width/height: {} != {}'.format(mask.shape[:2], image.shape[:2]))
    if mask.ndim == 2:
        if separate_signs:
            mask = split_signs(mask)
        else:
            mask = stack_label(mask, label_channel, as_uint8=False)
    elif mask.ndim == 3 and mask.shape[2] == 3:
        pass
    else:
        raise ValueError("Mask must have shape [x, y] or [x, y, z], not {}".format(mask.shape))
    if image.dtype == np.uint8:
        mask = to_uint8(mask)

    if output.ndim == 2:
        output = np.dstack([output, output, output])

    overlay = cv2.addWeighted(output, 0.5, mask, 0.5, 0)
    output = np.where(mask.sum(axis=2, keepdims=True) != 0, overlay, output)
    return output


def masked_lineup(image, label):
    """Return a list of image, masked_image, mask.

    Note
    ----
        Image will not be changed, but the masked_image and mask will be uint8 arrays with range [0, 255].

    Parameters
    ----------
        image: np array
            Image array with shape (x, y, channels)
        label: np array
            np array with shape (x, y) and range [0, 1]
        norm_type: int | None
            Type of normalization to undo when masking the image. If None, it will
            be inferred based on the value range of the array.
    """
    return [
        image,
        add_overlay(image, label),
        np.dstack([label * 255,
                   np.zeros_like(label),
                   np.zeros_like(label)]).astype(np.uint8)
    ]


def grabCut(image, labels, use_tresholds=True, thresholds=(0.2, 0.6, 0.8), iterations=2, clean=False):
    """Use the predicted mask and OpenCV's GrabCut algorithm to mask an image.

    Parameters
    ----------
    image : numpy.ndarray
        input image to mask
    labels : numpy.ndarray
        labels for background/foreground.
    use_thresholds: bool
        If true, thresholds the labels to find FG/BG components. Otherwise assumes that the values are [0, 2, 3, 1] defined by OpenCV for [BG, PR_BG, PR_FG, FG]
    thresholds : (int, int, int)
        Thesholds to divide [BG, PR_BG, PR_FG, FG] in the input labels
    iterations : int
        Number of GrabCut iterations to perform (the default is 2).
    clean : bool
        Should the GrabCut results be smoothed using :func:`~gouda.image.clean_grabCut_mask` (the default is False).

    Returns
    -------
    The GrabCut masked image and the generated mask : numpy.ndarray, numpy.ndarray

    """
    if labels.max() <= 1:
        mask = np.zeros_like(labels, dtype=np.uint8)
        mask += cv2.GC_BGD
        mask[labels > thresholds[0]] = cv2.GC_PR_BGD
        mask[labels > thresholds[1]] = cv2.GC_PR_FGD
        mask[labels > thresholds[2]] = cv2.GC_FGD
    else:
        mask = labels.astype(np.uint8)

    if cv2.GC_FGD not in mask and cv2.GC_PR_FGD not in mask:
        raise ValueError("Labels cannot all be background")
    elif cv2.GC_BGD not in mask and cv2.GC_PR_BGD not in mask:
        raise ValueError("Labels cannot all be foreground")

    bg, fg = np.zeros((1, 65), np.float64), np.zeros((1, 65), np.float64)
    cv2.grabCut(image, mask, None, bg, fg, iterations, cv2.GC_INIT_WITH_MASK)

    if clean:
        mask = clean_grabCut_mask(mask)
    mask_out = np.where((mask == 1) + (mask == 3), 1, 0).astype(np.uint8)
    masked_img = cv2.bitwise_and(image, image, mask=mask_out)
    return masked_img, mask_out


def clean_grabCut_mask(mask):
    """Apply light smoothing to grabCut labels"""
    temp = np.copy(mask)
    change = np.where((temp == 2) + (temp == 3), 1, 0)
    remapped = np.zeros_like(temp)
    remapped[temp == 0] = 0
    remapped[temp == 1] = 3
    remapped[temp == 2] = 1
    remapped[temp == 3] = 2
    up = cv2.pyrUp(remapped)
    for i in range(15):
        up = cv2.medianBlur(up, 7)
    down = cv2.pyrDown(up)
    down = np.round(down)
    if temp.ndim == 3:
        down = down[:, :, np.newaxis]
    remapped = np.where(change == 1, down, remapped)
    output = np.zeros_like(temp)
    output[remapped == 0] = 0
    output[remapped == 3] = 1
    output[remapped == 1] = 2
    output[remapped == 2] = 3
    return output.reshape(mask.shape).astype(mask.dtype)


def crop_to_mask(image, mask, with_label=False, smoothing=True):
    """Crop input image to only be size of input mask.

    Parameters
    ----------
        image: numpy uint8 array
            Image to mask/crop
        mask: numpy bool array
            Masking array with same x and y shape as image
        with_label: bool
            Whether or not to return a version of the mask that is cropped to the new boundaries
        smoothing: bool
            Whether or not to smooth the border of the mask

    Returns
    -------
        numpy uint8 array
            Masked image with dimensions that are only as large
            as the positive region of the mask
    """
    vert = np.any(mask, axis=0)
    y_range = np.where(vert == True)  # noqa: E712
    horiz = np.any(mask, axis=1)
    x_range = np.where(horiz == True)  # noqa: E712
    x0, x1 = x_range[0][0], x_range[0][-1]
    y0, y1 = y_range[0][0], y_range[0][-1]
    if smoothing:
        smooth_mask = cv2.GaussianBlur(
            mask.astype(np.uint8) * 255.0, (5, 5), 0)
        # mask[mask<0.8] = 0
        # mask[mask>=0.8] = 1
    else:
        smooth_mask = mask.astype(np.uint8) * 255.0
    masked_image = cv2.bitwise_and(image,
                                   image,
                                   mask=smooth_mask.astype(np.uint8))
    if with_label:
        return masked_image[x0:x1, y0:y1], mask[x0:x1, y0:y1]
    return masked_image[x0:x1, y0:y1]


def get_bounds(mask):
    """Return the bounding box corners of the positive content for the mask."""
    vert = np.any(mask, axis=0)
    horiz = np.any(mask, axis=1)
    y_range = np.where(vert == True)  # noqa: E712
    x_range = np.where(horiz == True)  # noqa: E712
    x0, x1 = x_range[0][0], x_range[0][-1]
    y0, y1 = y_range[0][0], y_range[0][-1]
    return (x0, y0), (x1, y1)


def crop_to_content(image):
    """Crop image to only be as large as the contained image excluding black space."""
    vert = np.mean(image, axis=(-1, 0))
    y_range = np.where(vert > 0)
    horiz = np.mean(image, axis=(-1, 1))
    x_range = np.where(horiz > 0)
    x0, x1 = x_range[0][0], x_range[0][-1]
    y0, y1 = y_range[0][0], y_range[0][-1]
    return image[x0:x1, y0:y1]


def rotate(img, degrees=90, allow_resize=True):
    """Rotate image clock-wise using OpenCV.

    Parameters
    ----------
    img: image (as required by OpenCV getRotationMatrix2D)
        image/path to rotate
    degrees: int
        Degrees to rotate the image clockwise
    allow_resize: bool
        If yes, the image boundaries will change to fit the rotated image. Otherwise the output rotated image is cropped to fit the original boundaries.
    """
    (h, w) = img.shape[:2]
    (cX, cY) = (w / 2, h / 2)
    M = cv2.getRotationMatrix2D((cX, cY), -degrees, 1.0)
    cos = np.abs(M[0, 0])
    sin = np.abs(M[0, 1])
    if allow_resize:
        nW = int((h * sin) + (w * cos))
        nH = int((h * cos) + (w * sin))
    else:
        nW, nH = w, h
    M[0, 2] += (nW / 2) - cX
    M[1, 2] += (nH / 2) - cY
    # Fixes a 1-pixel offset courtesy of: https://github.com/opencv/opencv/issues/4585#issuecomment-397895187
    M[0, 2] += (M[0, 0] + M[0, 1] - 1) / 2
    M[1, 2] += (M[1, 0] + M[1, 1] - 1) / 2
    return cv2.warpAffine(img, M, (nW, nH)).astype(img.dtype)


def padded_resize(image, size=[960, 540], allow_rotate=True):
    """Resize input image to given size, only padding as needed for aspect ratio.

    Parameters
    ----------
    image: image/path
        image/path to image to resize
    size: [int, int]
        output size
    allow_rotate: bool
        Whether the image can be rotated to minimize required padding or if orientation should be preserved
    Input image number of channels does not matter as long as the first two dimensions are x and y.
    """
    if type(image) == str:
        image = imread(image, as_RGB=True)
    data_type = image.dtype
    if image.ndim == 2:
        x, y = image.shape
        c = 0
        image = image[:, :, np.newaxis]
    else:
        x, y, c = image.shape
    if ((size[0] < size[1] and x > y) or (size[0] > size[1] and x < y)) and allow_rotate:
        image = rotate(image, 90)
        x, y = image.shape[:2]
    padx = int(y * (float(size[0]) / size[1])) - x
    pady = int(x * (float(size[1]) / size[0])) - y
    if abs(padx) > 10 or abs(pady) > 10:
        padx = 10000 if padx < 0 else padx
        pady = 10000 if pady < 0 else pady
        if padx < pady:
            new_shape = [x + padx, y, max(c, 1)]
            padded_image = np.zeros(new_shape, dtype=data_type)
            padded_image[int(padx // 2):-int(padx - padx // 2), :] = image
        else:
            new_shape = [x, y + pady, max(c, 1)]
            padded_image = np.zeros(new_shape, dtype=data_type)
            padded_image[:, int(pady // 2):-int(pady - pady // 2)] = image
    else:
        padded_image = image
    padded_image = cv2.resize(padded_image, (size[1], size[0])).astype(data_type)
    if c == 1:
        # cv2.resize auto-squeezes images
        return padded_image[:, :, np.newaxis]
    return padded_image


def horizontal_flip(image):
    """This is for convenience, but it is ~2e-7s faster to just copy the source and do this in-line"""
    return image[:, ::-1, :]


def vertical_flip(image):
    """This is for convenience, but it is ~1e-7s faster to just copy the source and do this in-line"""
    return image[::-1, ...]


def adjust_gamma(image, gamma=1.0):
    """Adjust the gamma of the image."""
    invGamma = 1.0 / gamma
    table = np.array([((i / 255.0)**invGamma) * 255
                      for i in np.arange(0, 256)]).astype(np.uint8)
    return cv2.LUT(image, table)


def polar_to_cartesian(image, output_shape=None):
    """Convert a square image with a polar object (circle/tube) to cartesian (unroll it).

    NOTE: output_shape uses numpy shape: [rows, columns]
    """
    if output_shape is None:
        output_shape = image.shape

    degrees = np.linspace(0, 360, output_shape[1], endpoint=False, dtype=np.float32)
    degrees = np.radians(degrees)
    center = image.shape[0] // 2
    radius = np.linspace(0, center, output_shape[0], endpoint=False, dtype=np.float32)
    d, r = np.meshgrid(degrees, radius)
    newx = r * np.cos(d) + center
    newy = r * np.sin(d) + center

    output = cv2.remap(image, newx, newy, cv2.INTER_LINEAR)
    return output



# -*- coding: utf-8 -*-
"""Miscellaneous data utilities."""
from pkg_resources import DistributionNotFound, get_distribution

from .binaryconfusionmatrix import BinaryConfusionMatrix  # noqa: F401
from .confusionmatrix import ConfusionMatrix  # noqa: F401
from .data_methods import *  # noqa: F401, F403
from .display import print_grid, print_image  # noqa: F401
from .file_methods import *  # noqa: F401, F403
from .goudapath import GoudaPath  # noqa: F401
from .moving_stats import *  # noqa: F401, F403

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = 'GOUDA'
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:  # pragma: no cover
    __version__ = 'unknown'
finally:
    del get_distribution, DistributionNotFound



import json
import os
import pathlib

import pytest

import numpy as np

import gouda

# import pytest


def test_ensure_dir():
    test_dir = gouda.ensure_dir("test_dir")
    test_dir2 = gouda.ensure_dir(gouda.GoudaPath("test_dir", use_absolute=False))
    assert test_dir == test_dir2.path
    assert os.path.isdir("test_dir")
    assert test_dir == "test_dir"

    test_dir_path = gouda.ensure_dir("test_dir", "check1")
    assert test_dir_path == os.path.join(test_dir, "check1")
    assert os.path.isdir(test_dir_path)

    pathlib.Path(os.path.join(test_dir_path, 'check2')).touch()
    with pytest.raises(ValueError):
        assert gouda.ensure_dir(test_dir_path, "check2")

    # Cleanup
    os.remove(os.path.join(test_dir_path, 'check2'))
    os.rmdir(test_dir_path)
    os.rmdir('test_dir')


def test_next_filename():
    assert gouda.next_filename("test.txt") == "test.txt"
    open('test.txt', 'w').close()
    assert gouda.next_filename("test.txt") == "test_2.txt"
    open("test_2.txt", 'w').close()
    assert gouda.next_filename("test.txt") == "test_3.txt"
    assert gouda.next_filename("test_2.txt") == "test_2_2.txt"
    # Cleanup
    os.remove("test.txt")
    os.remove("test_2.txt")


def test_save_load_json_dict():
    temp_data = {'a': 1, 'b': 2, 'c': 3}
    gouda.save_json(temp_data, 'test.json')
    assert os.path.isfile('test.json')
    check_data = gouda.load_json('test.json')
    for key in temp_data.keys():
        assert temp_data[key] == check_data[key]
    os.remove('test.json')


def test_save_load_json_list():
    temp_data = ['a', 'b', 'c']
    gouda.save_json(temp_data, 'test.json')
    assert os.path.isfile('test.json')
    check_data = gouda.load_json('test.json')
    for i in range(len(temp_data)):
        assert temp_data[i] == check_data[i]
    os.remove('test.json')


def test_save_load_json_nested():
    temp_data = [{'a': 1}, {'a': 2}, {'a': 3}]
    gouda.save_json(temp_data, 'test.json')
    assert os.path.isfile('test.json')
    check_data = gouda.load_json('test.json')
    for i in range(len(temp_data)):
        assert temp_data[i]['a'] == check_data[i]['a']
    os.remove('test.json')


def test_save_json_load_json_numpy():
    temp_data = np.arange(5, dtype=np.uint8)
    gouda.save_json(temp_data, 'testx.json', embed_arrays=True, compressed=False)
    gouda.save_json(temp_data, 'test2.json', embed_arrays=False, compressed=False)
    gouda.save_json(temp_data, 'test3.json', embed_arrays=False, compressed=True)

    assert os.path.isfile('testx.json')
    assert not os.path.isfile('testx_array.npz')
    assert not os.path.isfile('testx_arrayzip.npz')
    assert os.path.isfile('test2.json')
    assert os.path.isfile('test2_array.npz')
    assert os.path.isfile('test3.json')
    assert os.path.isfile('test3_arrayzip.npz')

    with open('testx.json', 'r') as f:
        data = json.load(f)
        assert data[-1] == 'numpy_embed'
    with open('test2.json', 'r') as f:
        data = json.load(f)
        assert data[-1] == 'numpy'
    with open('test3.json', 'r') as f:
        data = json.load(f)
        assert data[-1] == 'numpy_zip'

    check_data = gouda.load_json('testx.json')
    check_data2 = gouda.load_json('test2.json')
    check_data3 = gouda.load_json('test3.json')
    np.testing.assert_array_equal(temp_data, check_data)
    np.testing.assert_array_equal(temp_data, check_data2)
    np.testing.assert_array_equal(temp_data, check_data3)

    os.remove('testx.json')
    os.remove('test2.json')
    os.remove('test3.json')
    os.remove('test2_array.npz')
    os.remove('test3_arrayzip.npz')


def test_save_load_json_list_numpy():
    temp_data = np.arange(5, dtype=np.uint8)

    gouda.save_json([temp_data], 'test1.json', embed_arrays=False, compressed=False)
    gouda.save_json([temp_data], 'test2.json', embed_arrays=True, compressed=False)
    gouda.save_json([temp_data], 'test3.json', embed_arrays=False, compressed=True)

    check1 = gouda.load_json('test1.json')
    check2 = gouda.load_json('test2.json')
    check3 = gouda.load_json('test3.json')
    assert isinstance(check1, list)
    assert isinstance(check2, list)
    assert isinstance(check3, list)

    np.testing.assert_array_equal(temp_data, check1[0])
    np.testing.assert_array_equal(temp_data, check2[0])
    np.testing.assert_array_equal(temp_data, check3[0])


def test_save_load_json_numpy_list():
    temp_data = [np.arange(3, dtype=np.uint8), np.arange(4, 6, dtype=np.float32)]
    gouda.save_json(temp_data, 'test.json', embed_arrays=True, compressed=False)
    check_data = gouda.load_json('test.json')
    assert len(check_data) == 2
    np.testing.assert_array_equal(temp_data[0], check_data[0])
    np.testing.assert_array_equal(temp_data[1], check_data[1])

    gouda.save_json(temp_data, 'test2.json', embed_arrays=False, compressed=False)
    check_data2 = gouda.load_json('test2.json')
    assert len(check_data2) == 2
    np.testing.assert_array_equal(temp_data[0], check_data2[0])
    np.testing.assert_array_equal(temp_data[1], check_data2[1])

    os.remove('test.json')
    os.remove('test2.json')
    os.remove('test2_array.npz')


def test_save_load_json_numpy_dict():
    temp_data = {'a': np.arange(3, dtype=np.uint8)}
    gouda.save_json(temp_data, 'test.json', embed_arrays=True, compressed=False)
    check_data = gouda.load_json('test.json')
    assert len(temp_data) == len(check_data)
    np.testing.assert_array_equal(temp_data['a'], check_data['a'])

    gouda.save_json(temp_data, 'test2.json', embed_arrays=False, compressed=False)
    check_data2 = gouda.load_json('test2.json')
    assert len(temp_data) == len(check_data2)
    np.testing.assert_array_equal(temp_data['a'], check_data2['a'])

    os.remove('test.json')
    os.remove('test2.json')
    os.remove('test2_array.npz')


def test_save_load_json_numpy_mixed():
    temp_data = [np.arange(3), 3]
    gouda.save_json(temp_data, 'test.json', embed_arrays=True, compressed=False)
    check_data = gouda.load_json('test.json')
    np.testing.assert_array_equal(temp_data[0], check_data[0])
    assert check_data[1] == 3

    gouda.save_json(temp_data, 'test2.json', embed_arrays=False, compressed=False)
    check_data2 = gouda.load_json('test2.json')
    np.testing.assert_array_equal(temp_data[0], check_data2[0])
    assert check_data2[1] == 3

    os.remove('test.json')
    os.remove('test2.json')
    os.remove('test2_array.npz')


def test_save_json_warning_error():
    temp_data = [1, 2, 3]
    with pytest.warns(UserWarning):
        gouda.save_json(temp_data, 'test.json', embed_arrays=True, compressed=True)
    os.remove('test.json')

    temp_data = np.arange(5, dtype=np.uint8)
    gouda.save_json(temp_data, 'test.json', embed_arrays=False)
    gouda.save_json(temp_data.astype(np.float32), 'test2.json', embed_arrays=False)
    gouda.save_json(temp_data.reshape([1, 5]), 'test3.json', embed_arrays=False)
    os.remove('test_array.npz')
    os.rename('test2_array.npz', 'test_array.npz')
    with pytest.raises(ValueError):
        gouda.load_json('test.json')
    os.remove('test_array.npz')
    os.rename('test3_array.npz', 'test_array.npz')
    with pytest.raises(ValueError):
        gouda.load_json('test.json')

    os.remove('test.json')
    os.remove('test2.json')
    os.remove('test3.json')
    os.remove('test_array.npz')


def test_is_image():
    test_image = np.ones([20, 20])
    gouda.image.imwrite('test_image.png', test_image)
    assert gouda.is_image('test_image.png')
    assert gouda.is_image('.') is False
    with pytest.raises(FileNotFoundError):
        assert gouda.is_image('asdfhaksdfhklasjdhfakdhsfk.asdf')
    path = gouda.GoudaPath('test_image.png')
    assert gouda.is_image(path)

    os.remove('test_image.png')
    assert not os.path.exists('test_image.png')


def test_save_load_json_slice():
    data = slice(0, 10, None)
    gouda.save_json(data, 'test.json')
    compare = gouda.load_json('test.json')
    assert compare == data
    os.remove('test.json')

    data = slice(0, 10, 2)
    gouda.save_json(data, 'test.json')
    compare = gouda.load_json('test.json')
    assert compare == data
    os.remove('test.json')

    data = [slice(0, 10, 2)]
    gouda.save_json(data, 'test.json')
    compare = gouda.load_json('test.json')
    print(compare, data)
    assert compare == data
    os.remove('test.json')

    data = [slice(0, 10, 2), {'this': slice(100, 230, None), 'that': np.array([1, 2, 3])}]
    gouda.save_json(data, 'test.json')
    compare = gouda.load_json('test.json')
    assert compare[0] == slice(0, 10, 2)
    assert isinstance(compare[1], dict)
    assert compare[1]['this'] == slice(100, 230)
    np.testing.assert_array_equal(compare[1]['that'], np.array([1, 2, 3]))
    os.remove('test.json')



import os

import pytest

import matplotlib.pyplot as plt
import numpy as np

from gouda import display


def test_print_grid():
    # List Tests
    image = np.random.randint(0, 255, size=[100, 100])

    # Image
    assert display.print_grid(image, show=False, return_grid_shape=True) == (1, 1)
    plt.close()

    # # Multiple images
    assert display.print_grid(image, image, show=False, return_grid_shape=True) == (1, 2)
    plt.close()

    # # List of Images
    assert display.print_grid([image], show=False, return_grid_shape=True) == (1, 1)
    plt.close()
    assert display.print_grid([image, image], show=False, return_grid_shape=True) == (1, 2)
    plt.close()

    # # List of List of Images
    assert display.print_grid([image], show=False, return_grid_shape=True) == (1, 1)
    plt.close()
    assert display.print_grid([[image, image]], show=False, return_grid_shape=True) == (1, 2)
    plt.close()
    assert display.print_grid([[image], [image]], show=False, return_grid_shape=True) == (2, 1)
    plt.close()
    assert display.print_grid([[image, image], [image]], show=False, return_grid_shape=True) == (2, 2)
    plt.close()
    assert display.print_grid([[image, image], [None, image]], show=False, return_grid_shape=True) == (2, 2)
    plt.close()

    # Array Tests
    image_0c = np.random.randint(0, 255, size=[100, 100])
    image_1c = np.random.randint(0, 255, size=[100, 100, 1])
    image_3c = np.random.randint(0, 255, size=[100, 100, 3])

    # [x, y], [x, y, 1], [x, y, 3]
    assert display.print_grid(image_0c, show=False, return_grid_shape=True) == (1, 1)
    plt.close()
    assert display.print_grid(image_1c, show=False, return_grid_shape=True) == (1, 1)
    plt.close()
    assert display.print_grid(image_3c, show=False, return_grid_shape=True) == (1, 1)
    plt.close()

    # [cols, x, y], [cols, x, y, 1], [cols, x, y, 3]
    assert display.print_grid(image_0c[np.newaxis], show=False, return_grid_shape=True) == (1, 1)
    plt.close()
    assert display.print_grid(np.stack([image_0c, image_0c], axis=0), show=False, return_grid_shape=True) == (1, 2)
    plt.close()
    assert display.print_grid(image_1c[np.newaxis], show=False, return_grid_shape=True) == (1, 1)
    plt.close()
    assert display.print_grid(np.stack([image_1c, image_1c], axis=0), show=False, return_grid_shape=True) == (1, 2)
    plt.close()
    assert display.print_grid(image_3c[np.newaxis], show=False, return_grid_shape=True) == (1, 1)
    plt.close()
    assert display.print_grid(np.stack([image_3c, image_3c], axis=0), show=False, return_grid_shape=True) == (1, 2)
    plt.close()

    # [rows, cols, x, y], [rows, cols, x, y, 1], [rows, cols, x, y, 3]
    assert display.print_grid(image_0c[np.newaxis, np.newaxis], show=False, return_grid_shape=True) == (1, 1)
    plt.close()
    assert display.print_grid(np.stack([image_0c, image_0c], axis=0)[np.newaxis],
                      show=False, return_grid_shape=True) == (1, 2)
    plt.close()
    assert display.print_grid(np.stack([image_0c[np.newaxis], image_0c[np.newaxis]], axis=0),
                              show=False, return_grid_shape=True) == (2, 1)
    plt.close()
    assert display.print_grid(np.stack([
        np.stack([image_0c, image_0c], axis=0),
        np.stack([image_0c, image_0c], axis=0)]),
                              show=False, return_grid_shape=True) == (2, 2)
    plt.close()

    assert display.print_grid(image_1c[np.newaxis, np.newaxis],
                              show=False, return_grid_shape=True) == (1, 1)
    plt.close()
    assert display.print_grid(np.stack([image_1c, image_1c], axis=0)[np.newaxis],
                              show=False, return_grid_shape=True) == (1, 2)
    plt.close()
    assert display.print_grid(np.stack([image_1c[np.newaxis], image_1c[np.newaxis]], axis=0),
                              show=False, return_grid_shape=True) == (2, 1)
    plt.close()
    assert display.print_grid(np.stack([
        np.stack([image_1c, image_1c], axis=0),
        np.stack([image_1c, image_1c], axis=0)]),
                              show=False, return_grid_shape=True) == (2, 2)
    plt.close()

    assert display.print_grid(image_3c[np.newaxis, np.newaxis],
                              show=False, return_grid_shape=True) == (1, 1)
    plt.close()
    assert display.print_grid(np.stack([image_3c, image_3c], axis=0)[np.newaxis],
                              show=False, return_grid_shape=True) == (1, 2)
    plt.close()
    assert display.print_grid(np.stack([image_3c[np.newaxis], image_3c[np.newaxis]], axis=0),
                              show=False, return_grid_shape=True) == (2, 1)
    plt.close()
    assert display.print_grid(np.stack([
        np.stack([image_3c, image_3c], axis=0),
        np.stack([image_3c, image_3c], axis=0)]),
                              show=False, return_grid_shape=True) == (2, 2)
    plt.close()

    test_image = {'image': image_3c}
    assert display.print_grid(test_image, show=False, return_grid_shape=True, suptitle='test_sup') == (1, 1)
    test_image = {'image': image_1c}
    assert display.print_grid(test_image, show=False, return_grid_shape=True) == (1, 1)
    test_image = {'image': image_1c, 'cmap': 'viridis', 'title': 'test', 'xlabel': 'testx', 'ylabel': 'testy'}
    assert display.print_grid(test_image, show=False, return_grid_shape=True, left=0.1) == (1, 1)
    plt.close()


def test_print_grid_tofile():
    test_image = np.ones([10, 10])
    display.print_grid(test_image, show=False, toFile='testprintgrid.png')
    assert os.path.exists('testprintgrid.png')
    os.remove('testprintgrid.png')
    assert not os.path.exists('testprintgrid.png')


def test_print_grid_exceptions():
    test_image = np.ones([5, 5, 100, 100, 5])
    with pytest.raises(ValueError):
        assert display.print_grid(test_image)

    with pytest.raises(ValueError):
        assert display.print_grid('string')


def test_print_image():
    test_image = np.ones([100, 100])
    test_image[25:75, 25:75] = 0
    display.print_image(test_image, show=False, left=0.1)
    plt.close()
    display.print_image(test_image, show=False)
    plt.close()

    test_image = np.ones([100, 100, 3], dtype=np.uint8) * 255
    test_image[25:75, 25:75] = 0
    display.print_image(test_image, toFile='test_image.png', show=False)
    plt.close()
    assert os.path.exists('test_image.png')
    os.remove('test_image.png')
    assert not os.path.exists('test_image.png')
